from __future__ import print_function, division, absolute_import

import numpy as np
import tensorflow as tf

from odin import backend as K
from odin.nnet.base import NNOp
from odin.utils import is_callable, is_pickleable

_SUPPORT_REDUCE_FUNCTION = [
    tf.reduce_sum,
    tf.reduce_mean,
    tf.reduce_max,
    tf.reduce_min,
    tf.reduce_prod,
    tf.reduce_logsumexp,
    tf.reduce_all,
    tf.reduce_any
]

class Reduce(NNOp):
  """ support all Reduce function in tensorflow"""

  def __init__(self, fn, axis=None, keepdims=False, **kwargs):
    super(Reduce, self).__init__(**kwargs)
    if not is_callable(fn):
      raise ValueError("`fn` must be callable")
    if fn not in _SUPPORT_REDUCE_FUNCTION:
      raise ValueError("Not support for function type: %s, all support function include: %s" %
        (str(fn), ', '.join([i.__name__ for i in _SUPPORT_REDUCE_FUNCTION])))
    self.fn = fn
    self.axis = axis
    self.keepdims = bool(keepdims)

  def _transpose(self):
    raise NotImplementedError

  def _apply(self, X):
    return self.fn(X, axis=self.axis, keepdims=self.keepdims)

class Repeat(NNOp):
  """ Repeat a dimension of a tensor

  If x has shape (s1, s2, s3) and axes=(1, -1), the output
  will have shape (s1, s2 * n[0], s3 * n[1]).

  Parameters
  ----------
  n : {int, list of int}
    each number of repeatation according to the axes
  axes : {int, list or int}
    all axes for repeating
  """

  def __init__(self, n, axes):
    super(Repeat, self).__init__()
    self.n = n
    self.axes = axes

  def _transpose(self):
    raise NotImplementedError

  def _apply(self, X):
    return K.repeat(X, n=self.n, axes=self.axes)

class Activate(NNOp):
  """ Custom activation function
  """

  def __init__(self, fn, keywords={}, **kwargs):
    super(Activate, self).__init__(**kwargs)
    if not is_callable(fn):
      raise ValueError("`fn` must be callable")
    if not is_pickleable(fn):
      raise ValueError("`fn` must be pickle-able")
    self.fn = fn
    self.keywords = keywords

  def _transpose(self):
    return Activate(self.fn, keywords=self.keywords)

  def _apply(self, X):
    return self.fn(X, **self.keywords)



# -*- coding: utf-8 -*-
"""
This module contain all pretrained Model, which include:

- BNF_*: pretrained deep bottleneck features based on
    acoustic features
- VGG_*: collection of different VGG architecture for image
    recogition

Example
-------
  >>> import numpy as np
  >>> import tensorflow as tf
  ...
  >>> from odin import nnet as N
  ...
  >>> dbf = N.models.BNF_1024_MFCC39()
  >>> print(dbf.get_input_info())
  ... # {'X': ((None, 819), 'float32')}
  ... # name -> (shape, dtype):
  >>> X = np.random.rand(128, 819).astype('float32')
  ...
  >>> with tf.device('/cpu:0'):
  ...     y_cpu = dbf(X)
  ...
  >>> with tf.device('/gpu:0'):
  ...     y_gpu = dbf(X)
  ...
  >>> assert np.all(np.isclose(y_cpu, y_gpu))

"""

from odin.nnet.models.base import Model
from odin.nnet.models.bnf import *
from odin.nnet.models.imagenet import *



from odin.ml.base import evaluate
from odin.ml.decompositions import *
from odin.ml.gmm_tmat import GMM, Tmatrix
from odin.ml.ivector import Ivector
from odin.ml.linear_model import LogisticRegression
from odin.ml.deep_model import *
from odin.ml.scoring import (VectorNormalizer, Scorer,
                      compute_wccn, compute_class_avg, compute_within_cov)
from odin.ml.plda import PLDA
from odin.ml.gmm_classifier import GMMclassifier
from odin.ml.fast_tsne import fast_tsne



from odin.fuel.utils import *
from odin.fuel.data import *
from odin.fuel.dataset import *
from odin.fuel.feeder import *
from odin.fuel.loaders import *
from odin.fuel.synthetic_loaders import *

from odin.fuel import recipes



# -*- coding: utf-8 -*-
from __future__ import print_function
import os
import io
import sys
# import uuid
import time
import math
import types
import signal
import shutil
import timeit
import inspect
import tarfile
import numbers
import tempfile
import platform
import argparse
import subprocess
import contextlib
from multiprocessing import cpu_count, Lock, current_process
from collections import OrderedDict, deque, Iterable, Iterator, Mapping
from itertools import islice, tee, chain

from six import string_types
from six.moves import cPickle
from six.moves.urllib.request import urlopen
from six.moves.urllib.error import URLError, HTTPError

try:
  from numba import jit, autojit, vectorize, guvectorize
except:
  pass
import numpy

from odin.utils.progbar import Progbar, add_notification
from odin.utils.path_utils import *
from odin.utils.cache_utils import *
from odin.utils.python_utils import *
from odin.utils.np_utils import *
from odin.utils.mpi import segment_list, SharedCounter, async, async_mpi, MPI
from odin.utils.crypto import md5_checksum

from odin.utils import mpi
from odin.utils import shape_calculation
from odin.utils import decorators
from odin.utils import crypto

def array_size(arr):
  """ Return size of an numpy.ndarray in bytes """
  return np.prod(arr.shape) * arr.dtype.itemsize

# ===========================================================================
# Pretty print
# ===========================================================================
def ctext(s, color='red'):
  """ Colored text support
  * BLACK
  * RED
  * GREEN
  * YELLOW
  * BLUE
  * MAGENTA
  * CYAN
  * WHITE
  * RESET
  * LIGHTBLACK_EX
  * LIGHTRED_EX
  * LIGHTGREEN_EX
  * LIGHTYELLOW_EX
  * LIGHTBLUE_EX
  * LIGHTMAGENTA_EX
  * LIGHTCYAN_EX
  * LIGHTWHITE_EX

  Note
  ----
  * colored text is longer in length and can overlength the screen
  (i.e. make the progress-bar ugly).
  * ctext(x + y + z) is BETTER than ctext(x) + ctext(y) + ctext(z)
  because it save more space.
  """
  s = str(s)
  try:
    from colorama import Fore
    color = str(color).strip()
    color = color.upper()
    # special fix for light color
    if 'LIGHT' == color[:5] and '_EX' != color[-3:]:
      color = color + '_EX'
    color = getattr(Fore, color, '')
    return color + s + Fore.RESET
  except ImportError:
    pass
  return s


def _type_name(x):
  if isinstance(x, np.dtype):
    return x.name
  return type(x).__name__


def dummy_formatter(x):
  s = str(x)
  if len(s) < 120 and '\n' not in s:
    return s
  if isinstance(x, (tuple, list)):
    return "(list)length=%d;type=%s" % \
        (len(x), _type_name(x[0]) if len(x) > 0 else '*empty*')
  if isinstance(x, np.ndarray):
    return "(ndarray)shape=%s;dtype=%s" % (str(x.shape), str(x.dtype))
  if isinstance(x, Mapping):
    return "(map)length=%d;dtype=%s" % (len(x),
        ';'.join([_type_name(i) for i in next(iter(x.items()))]))
  # dataset type
  if 'odin.fuel.dataset.Dataset' in str(type(x)):
    return ("(ds)path:%s" % x.path)
  # NNOp types
  if any('odin.nnet.base.NNOp' in str(i) for i in type.mro(type(x))):
    return "\n" + '\n'.join(['\t' + line for line in str(x).split('\n')])
  if is_string(x):
    return str(x) if len(x) < 250 else '(str)length:%d' % len(x)
  return str(x)


@contextlib.contextmanager
def UnitTimer(factor=1, name=None):
  start = timeit.default_timer()
  yield None
  end = timeit.default_timer()
  # set name for timing task
  if name is None:
    name = 'Task'
  print('"%s"' % ctext(name, 'yellow'),
      "Time:",
      ctext((end - start) / factor, 'cyan'),
      '(sec)')


# ===========================================================================
# Basics
# ===========================================================================
def is_same_shape(shape1, shape2):
  """
  Return
  ------
  True if two objects is the same shape tuple,
  otherwise, False
  """
  # if is string, evaluate to python object
  if is_string(shape1):
    shape1 = eval(shape1)
  if is_string(shape2):
    shape2 = eval(shape2)
  # tensorflow TensorShape
  if hasattr(shape1, 'as_list'):
    shape1 = shape1.as_list()
  if hasattr(shape2, 'as_list'):
    shape2 = shape2.as_list()
  # if is number, convert to shape tuple
  if is_number(shape1) or shape1 is None:
    shape1 = (shape1,)
  shape1 = as_tuple(shape1)
  if is_number(shape2) or shape2 is None:
    shape2 = (shape2,)
  shape2 = as_tuple(shape2)
  # check the number of dimension
  if len(shape1) != len(shape2):
    return False
  for s1, s2 in zip(shape1, shape2):
    if s1 is not None and s2 is not None and s1 != s2:
      return False
  return True

def is_fileobj(f):
  """ Check if an object `f` is intance of FileIO object created
  by `open()`"""
  return isinstance(f, io.TextIOBase) or \
      isinstance(f, io.BufferedIOBase) or \
      isinstance(f, io.RawIOBase) or \
      isinstance(f, io.IOBase)

def is_callable(x):
  return hasattr(x, '__call__')

def is_string(s):
  return isinstance(s, string_types)

def is_path(path):
  if is_string(path):
    try:
      os.path.exists(path)
      return True
    except Exception as e:
      return False
  return False


def is_number(i):
  return isinstance(i, numbers.Number)


def is_bool(b):
  return isinstance(b, type(True))


def is_primitives(x, inc_ndarray=True, exception_types=[]):
  """Primitive types include: number, string, boolean, None
  and numpy.ndarray (optional) and numpy.generic (optional)

  Parameters
  ----------
  inc_ndarray: bool
      if True, include `numpy.ndarray` and `numpy.generic` as a primitive types
  """
  # complex list or Mapping
  if isinstance(x, (tuple, list)):
    return all(is_primitives(i, inc_ndarray=inc_ndarray,
                             exception_types=exception_types)
               for i in x)
  elif isinstance(x, Mapping):
    return all(is_primitives(i, inc_ndarray=inc_ndarray,
                             exception_types=exception_types) and
               is_primitives(j, inc_ndarray=inc_ndarray,
                             exception_types=exception_types)
               for i, j in x.items())
  # check for number, string, bool, and numpy array
  if is_number(x) or is_string(x) or is_bool(x) or x is None or \
  (any(isinstance(x, t) for t in exception_types)) or \
  (inc_ndarray and isinstance(x, (numpy.ndarray, numpy.generic))):
    return True
  return False


def type_path(obj):
  clazz = type(obj)
  return clazz.__module__ + "." + clazz.__name__


def is_lambda(v):
  LAMBDA = lambda: 0
  return isinstance(v, type(LAMBDA)) and v.__name__ == LAMBDA.__name__


def is_pickleable(x):
  try:
    cPickle.dumps(x, protocol=cPickle.HIGHEST_PROTOCOL)
    return True
  except cPickle.PickleError:
    return False


def iter_chunk(it, n):
  """ Chunking an iterator into small chunk of size `n`
  Note: this can be used to slice data into mini batches
  """
  if not isinstance(it, Iterator):
    it = iter(it)
  obj = list(islice(it, n))
  while obj:
    yield obj
    obj = list(islice(it, n))


def to_bytes(x, nbytes=None, order='little'):
  """ Convert some python object to bytes array, support type:
  * string, unicode
  * integer
  * numpy.ndarray

  Note
  ----
  This method is SLOW
  """
  if is_string(x):
    return x.encode()
  elif isinstance(x, int):
    return x.to_bytes(nbytes, order, signed=False)
  elif isinstance(x, np.ndarray):
    return x.tobytes()
  else:
    raise ValueError("Not support bytes conversion for type: %s" %
        type(x).__name__)


def batching(batch_size, n=None, start=0, end=None, seed=None):
  """
  Parameters
  ----------
  batch_size: int
      number of samples for 1 single batch.
  n: {int, None}
      number of samples
  start : int (default: 0)
      starting point of the iteration (in number of sample)
  end : {int, None}
      ending point of the iteration (in number of sample),
      end = total_amount_of_sample if None
  seed : {int, None}
      random seed for shuffling the returned batches,
      no shuffling performed if seed is None

  Return
  ------
  iteration: [(start, end), (start, end), ...]
  """
  if end is None and n is None:
    raise ValueError('you must provide either `end` or `n`')
  if end is None:
    end = n
  start = int(start)
  end = int(end)
  assert end > start, "`end` must > `start`"
  batch_size = int(batch_size)
  # ====== no shuffling ====== #
  if seed is None:
    return ((i, min(i + batch_size, end))
            for i in range(start, end + batch_size, batch_size)
            if i < end)
  batches = list(range(start, end + batch_size, batch_size))
  rand = np.random.RandomState(seed)
  rand.shuffle(batches)
  return (((i, min(i + batch_size, end)))
          for i in batches
          if i < end)

def read_lines(file_path):
  if not os.path.exists(file_path):
    raise ValueError('File at path: %s does not exist.' % file_path)
  if os.path.isdir(file_path):
    raise ValueError("Path to %s is a folder" % file_path)
  lines = []
  with open(file_path, 'r') as f:
    for i in f:
      lines.append(i[:-1] if i[-1] == '\n' else i)
  return lines


# ===========================================================================
# Others
# ===========================================================================
def raise_return(e):
  raise e


_CURRENT_STDIO = None


class _LogWrapper():

  def __init__(self, stream, own_buffer):
    self.stream = stream
    self.own_buffer = bool(own_buffer)

  def write(self, message):
    # no backtrack for writing to file
    self.stream.write(message.replace("\b", ''))
    sys.__stdout__.write(message)

  def flush(self):
    self.stream.flush()
    sys.__stdout__.flush()

  def end(self):
    self.stream.flush()
    sys.__stdout__.flush()
    if self.own_buffer:
      try:
        self.stream.close()
      except Exception:
        pass

def get_stdio_path():
  return _CURRENT_STDIO

def stdio(path=None, suppress=False, stderr=True):
  """
  Parameters
  ----------
  path: {None, str, io.StringIO}
      if str or StringIO, specified path for saving all stdout (and stderr)
      if None, reset to the system default stdout and stderr
  suppress: boolean
      totally turn-off all stdout (and stdeer)
  stderr:
      apply output file with stderr also

  NOTE
  ----
  Redirect the system logging can slightly decrease the
  performance in logging intensive application.
  """
  # turn off stdio
  if suppress:
    f = open(os.devnull, "w")
    path = None
  # reset
  elif path is None:
    f = None
  # redirect to a file
  elif is_string(path):
    f = _LogWrapper(open(path, "w"), own_buffer=True)
  # redirect to a buffer
  elif is_fileobj(path):
    f = _LogWrapper(path, own_buffer=False)
    if hasattr(path, 'name'):
      path = f.name
    else:
      path = str(path)
  else:
    raise ValueError("Unsupport for path=`%s` in `stdio`." %
        str(type(path)))
  # ====== set current stdio path ====== #
  global _CURRENT_STDIO
  _CURRENT_STDIO = path
  # ====== assign stdio ====== #
  if f is None: # reset
    if isinstance(sys.stdout, _LogWrapper):
      sys.stdout.end()
    if isinstance(sys.stderr, _LogWrapper):
      sys.stderr.end()
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__
  else: # redirect to file
    sys.stdout = f
    if stderr:
      sys.stderr = f

def auto_logging(log_dir=None, prefix='', num_max=None):
  """
  Parameters
  ----------
  log_dir : str
    path to directory stored all the logs
  prefix : str
    prefix to all saved log (using script name by default)
  num_max : {None, int}
    maximum number of log will be stored
  """
  if log_dir is None:
    log_dir = get_script_path()
  if not os.path.isdir(log_dir):
    raise ValueError("'%s' is not a directory" % str(log_dir))
  prefix = get_script_name() if prefix is None or len(prefix) == 0 else str(prefix)
  date_time = get_formatted_datetime(only_number=False)
  path = os.path.join(log_dir, prefix + '[%s]' % date_time + '.txt')
  # ====== check maximum number of log ====== #
  if num_max is not None:
    from stat import ST_CTIME
    num_max = int(num_max)
    past_logs = [i for i in os.listdir(log_dir) if prefix in i]
    past_logs = sorted(past_logs,
                       key=lambda x: os.stat(os.path.join(log_dir, x))[ST_CTIME],
                       reverse=True)
    # remove previous logs
    for i, name in enumerate(past_logs):
      if i >= num_max - 1:
        os.remove(os.path.join(log_dir, name))
  return stdio(path=path, suppress=False, stderr=True)

def eprint(text):
  """Print ERROR message to stderr"""
  print(ctext('[Error]', 'red') + str(text), file=sys.stderr)

def wprint(text):
  """Print WARNING message to stderr"""
  print(ctext('[Warning]', 'yellow') + str(text), file=sys.stderr)

# ===========================================================================
# Universal ID
# ===========================================================================
_uuid_chars = list(chain(map(chr, range(65, 91)),  # ABCD
                         map(chr, range(97, 123)),  # abcd
                         map(chr, range(48, 57)))) # 0123
_uuid_random_state = numpy.random.RandomState(int(str(int(time.time() * 100))[3:]))


def uuid(length=8):
  """ Generate random UUID 8 characters with very very low collision """
  # m = time.time()
  # uniqid = '%8x%4x' % (int(m), (m - int(m)) * 1000000)
  # uniqid = str(uuid.uuid4())[:8]
  uniquid = ''.join(
      _uuid_random_state.choice(_uuid_chars,
                                size=length,
                                replace=True))
  return uniquid


@contextlib.contextmanager
def change_recursion_limit(limit):
  """Temporarily changes the recursion limit."""
  old_limit = sys.getrecursionlimit()
  if old_limit < limit:
    sys.setrecursionlimit(limit)
  yield
  sys.setrecursionlimit(old_limit)


@contextlib.contextmanager
def signal_handling(sigint=None, sigtstp=None, sigquit=None):
  # We cannot handle SIGTERM, because it prevent subproces from
  # .terminate()
  orig_int = signal.getsignal(signal.SIGINT)
  orig_tstp = signal.getsignal(signal.SIGTSTP)
  orig_quit = signal.getsignal(signal.SIGQUIT)

  if sigint is not None: signal.signal(signal.SIGINT, sigint)
  if sigtstp is not None: signal.signal(signal.SIGTSTP, sigtstp)
  if sigquit is not None: signal.signal(signal.SIGQUIT, sigquit)

  yield
  # reset
  signal.signal(signal.SIGINT, orig_int)
  signal.signal(signal.SIGTSTP, orig_tstp)
  signal.signal(signal.SIGQUIT, orig_quit)


# ===========================================================================
# UniqueHasher
# ===========================================================================
class UniqueHasher(object):
  """ This hash create strictly unique hash value by using
  its memory to remember which key has been assigned

  Note
  ----
  This function use deterministic hash, which give the same
  id for all labels, whenever you call it
  """

  def __init__(self, nb_labels=None):
    super(UniqueHasher, self).__init__()
    self.nb_labels = nb_labels
    self._memory = {} # map: key -> hash_key
    self._current_hash = {} # map: hash_key -> key

  def hash(self, value):
    key = abs(hash(value))
    if self.nb_labels is not None:
      key = key % self.nb_labels
    # already processed hash
    if value in self._current_hash:
      return self._current_hash[value]
    # not yet processed
    if key in self._memory:
      if self.nb_labels is not None and \
          len(self._memory) >= self.nb_labels:
        raise Exception('All %d labels have been assigned, outbound value:"%s"' %
                        (self.nb_labels, value))
      else:
        while key in self._memory:
          key += 1
          if self.nb_labels is not None and key >= self.nb_labels:
            key = 0
    # key not in memory
    self._current_hash[value] = key
    self._memory[key] = value
    return key

  def __call__(self, value):
    return self.hash(value)

  def map(self, order, array):
    """ Re-order an ndarray to new column order """
    order = as_tuple(order)
    # get current order
    curr_order = self._current_hash.items()
    curr_order.sort(key=lambda x: x[1])
    curr_order = [i[0] for i in curr_order]
    # new column order
    order = [curr_order.index(i) for i in order]
    return array[:, order]


class FuncDesc(object):
  """ This class store the description of arguments given a function.
  Automatically match the argument and keyword-arguments in following
  order:
   - Match all positional arguments.
   - if the function has `varargs`, keep the remains positional arguments.
   - if the function has `keywords`, keep the unknown argument name in `kwargs`
   - Add the missing default kwargs.
   - Return new tuple of (arg, kwargs)
  """

  def __init__(self, func):
    super(FuncDesc, self).__init__()
    # copy information from other FuncDesc
    if isinstance(func, FuncDesc):
      self._args = func._args
      self._defaults = func._defaults
      self._is_include_args = func._inc_args
      self._is_include_kwargs = func._inc_kwargs
      self._func = func._func
    # extract information from a function or method
    elif inspect.isfunction(func) or inspect.ismethod(func) or \
    isinstance(func, decorators.functionable):
      if isinstance(func, decorators.functionable):
        sign = inspect.signature(func.function)
      else:
        sign = inspect.signature(func)
      self._args = [n for n, p in sign.parameters.items()
                    if p.kind not in (inspect.Parameter.VAR_POSITIONAL,
                                      inspect.Parameter.VAR_KEYWORD)]
      self._is_include_args = any(i.kind == inspect.Parameter.VAR_POSITIONAL
                           for i in sign.parameters.values())
      self._is_include_kwargs = any(i.kind == inspect.Parameter.VAR_KEYWORD
                             for i in sign.parameters.values())
      self._defaults = {n: p.default
                        for n, p in sign.parameters.items()
                        if p.default != inspect.Parameter.empty}
      self._func = func
    else:
      raise ValueError("`func` must be function, method, or FuncDesc, "
                       "but given: %s" % str(type(func)))
    self.__name__ = self._func.__name__

  def __getstate__(self):
    import dill
    # check if need to re-load function module during setstate
    if not is_lambda(self._func):
      module = inspect.getmodule(self._func)
      func_module_key = [i
                         for i, j in list(sys.modules.items())
                         if j == module][0]
      func_module_name = module.__name__
      func_module_path = inspect.getfile(module)
    else:
      func_module_key = None
      func_module_name = None
      func_module_path = None
    # using dill to dump function
    func_str = dill.dumps(self._func)
    return (func_str,
            func_module_key, func_module_name, func_module_path,
            self.__name__,
            self._args, self._is_include_args,
            self._is_include_kwargs, self._defaults)

  def __setstate__(self, states):
    import dill
    (func_str,
     func_module_key, func_module_name, func_module_path,
     self.__name__,
     self._args, self._is_include_args,
     self._is_include_kwargs, self._defaults) = states
    # iterate through all loaded modules to find module contain given function
    if func_module_key is not None and \
    func_module_name is not None and \
    func_module_path is not None:

      found = None
      for name, module in list(sys.modules.items()):
        try:
          if inspect.getfile(module) == func_module_path or \
          (module.__name__.split('.')[-1] == func_module_name and
           self.__name__ in dir(module)):
            found = (name, module)
        except Exception as e:
          pass

      if found is None:
        # TODO: search for module in sys.path
        import imp
        path = func_module_path
        script_path = get_script_path()
        folders = func_module_path.split('/')
        for i in range(1, len(folders)):
          if os.path.exists(path) and os.path.isfile(path):
            break
          path = os.path.join(script_path, '/'.join(folders[-i:]))
        imp.load_source(name=func_module_name, pathname=path)
      elif found[0] != func_module_name:
        sys.modules[func_module_name] = sys.modules[found[0]]

    # load function
    self._func = dill.loads(func_str)

  @property
  def name(self):
    return self.__name__

  @property
  def qualname(self):
    """ The qualified name of the class, function, method, descriptor,
    or generator instance.

    New in version 3.3.
    """
    assert hasattr(self._func, '__qualname__'), \
    "The attribute `__qualname__` only be introduced from python 3.3"
    return self._func.__qualname__

  @property
  def path(self):
    """ Return the absolute path to the script contain the function """
    return inspect.getfile(inspect.getmodule(self._func))

  @property
  def args(self):
    return tuple(self._args)

  @property
  def defaults(self):
    return dict(self._defaults)

  @property
  def is_args(self):
    return self._is_include_args

  @property
  def is_kwargs(self):
    return self._is_include_kwargs

  def _match(self, *args, **kwargs):
    keywords = OrderedDict()
    for name, val in zip(self.args, args):
      keywords[name] = val
    # extra args
    if self._is_include_args and len(self.args) < len(args):
      args = args[len(self.args):]
    else:
      args = ()
    # remove extra kwargs in inc_kwargs=False
    if not self._is_include_kwargs:
      kwargs = {name: kwargs[name] for name in self._args
                if name in kwargs}
    keywords.update(kwargs)
    # ====== update the default ====== #
    for name, val in self._defaults.items():
      if name not in keywords:
        keywords[name] = val
    return args, keywords

  def __call__(self, *args, **kwargs):
    args, kwargs = self._match(*args, **kwargs)
    return self._func(*args, **kwargs)

  def __str__(self):
    s = "<%s>args:%s defaults:%s varargs:%s keywords:%s" % \
        (ctext(self._func.__name__, 'cyan'), self._args, self._defaults,
            self._is_include_args, self._is_include_kwargs)
    return s

# ===========================================================================
# ArgCOntrol
# ===========================================================================
def args_parse(descriptions):
  """ Shortcut for parsing the argument from terminal
  command, by a list of following tuple:
    [(name, help, enum_values, default_value), ...]

  Parameter
  ---------
  descriptions : tuple, list
    a list of tuple contain 1 to 3 of following information
    (name, help, enum_values, default_value), where:
    - name (string) is the name of argument can be `name` positional,
      `-name` optional, or `--name` flag trigger
    - help (string) is text for printing help
    - enum_values (tuple, list, None) list of string (or Number) contain
      acceptable values (set to None if no values given)
    - default_value (Number, string) default value for the argument

  Note
  ----
  The order is different from `ArgController.add`
  """
  args = ArgController()
  for desc in descriptions:
    if len(desc) == 1:
      args.add(name=desc[0], help='')
    elif len(desc) == 2:
      args.add(name=desc[0], help=desc[1])
    elif len(desc) == 3:
      args.add(name=desc[0], help=desc[1], enum=desc[2])
    elif len(desc) == 4:
      args.add(name=desc[0], help=desc[1], enum=desc[2], default=desc[3])
    else:
      raise ValueError("No support for given description: %s" % str(desc))
  return args.parse()

class ArgController(object):
  """ Simple interface to argparse """

  def __init__(self, print_parsed=True):
    super(ArgController, self).__init__()
    self.parser = None
    self._require_input = False
    self.args_preprocessor = {}
    self.args_enum = {}
    self.name = []
    self.print_parsed = print_parsed

  def _process_name(self, name):
    if '--' == name[:2]:
      name = name[2:]
    elif '-' == name[0]:
      name = name[1:]
    name = name.replace('-', '_')
    return name

  def _is_positional(self, name):
    if name[0] != '-' and name[:2] != '--':
      return True
    return False

  def _parse_input(self, key, val):
    key = self._process_name(key)
    # ====== search if manual preprocessing available ====== #
    for i, preprocess in self.args_preprocessor.items():
      if key == i and preprocess is not None:
        return preprocess(val)
    # ====== auto preprocess ====== #
    try:
      val = float(val)
      if int(val) == val:
        val = int(val)
    except Exception as e:
      val = str(val)
    return val

  def add(self, name, help, default=None, enum=None,
          preprocess=None):
    """ NOTE: if the default value is not given, the argument is
    required

    Parameters
    ----------
    name: str
        [-name] for optional parameters
        [name] for positional parameters
    help: str
        description of the argument
    default: {str, Number}
        default value for the argument
    enum : {tuple, list}
        list or tuple of acceptant values, for post-process
        validating
    preprocess: call-able
        a function to perform preprocessing of the parsed argument
    """
    if self.parser is None:
      self.parser = argparse.ArgumentParser(
          description='Automatic argument parser (yes,true > True; no,false > False)',
          add_help=True)
    # ====== NO default value ====== #
    if default is None:
      if self._is_positional(name):
        self.parser.add_argument(name, help=help, type=str, action="store",
            metavar='')
      else:
        self.parser.add_argument(name, help=help, type=str, action="store",
            required=True, metavar='')
      self._require_input = True
    # ====== boolean default value ====== #
    elif isinstance(default, bool):
      help += ' (default: %s)' % str(default)
      self.parser.add_argument(name, help=help,
                               action="store_%s" % str(not default).lower())
      preprocess = lambda x: bool(x)
    # ====== add defaults value ====== #
    else:
      help += ' (default: %s)' % str(default)
      self.parser.add_argument(name, help=help, type=str, action="store",
          default=str(default), metavar='')

    # ====== store preprocess dictionary ====== #
    name = self._process_name(name)
    if not hasattr(preprocess, '__call__'):
      preprocess = None
    self.args_preprocessor[name] = preprocess
    # store the enum values
    if enum is not None and not isinstance(enum, (tuple, list)):
      enum = (enum,)
    self.args_enum[name] = enum
    return self

  def parse(self, config_path=None):
    if self.parser is None:
      raise Exception('Call add to assign at least 1 argument for '
                      'for the function.')
    exit_now = False
    # ====== validate arguments ====== #
    try:
      if len(sys.argv) == 1 and self._require_input:
        self.parser.print_help()
        exit_now = True
      else:
        args = self.parser.parse_args()
    except Exception as e:
      # if specfy version or help, don't need to print anything else
      if all(i not in ['-h', '--help', '-v', '--version']
             for i in sys.argv):
        self.parser.print_help()
      exit_now = True
    if exit_now:
      exit()
    # ====== parse the arguments ====== #
    try:
      args = {i: self._parse_input(i, j)
              for i, j in args._get_kwargs()}
    except Exception as e:
      print('Error parsing argument with name "%s"' % str(e))
      self.parser.print_help()
      exit()
    # ====== checking enumerate values ====== #
    for name, val in args.items():
      enum = self.args_enum[self._process_name(name)]
      if enum is None:
        continue
      if val not in enum:
        raise ValueError("Argument with name '%s' given value: `%s`, but "
                         "only accept one of the following: %s" %
                         (name, val, enum))
    # ====== reset everything ====== #
    self.parser = None
    self.args_preprocessor = {}
    self.args_enum = {}
    self._require_input = False
    # ====== print parsed values ====== #
    if self.print_parsed:
      max_len = max(len(i) for i in args.keys())
      max_len = '%-' + str(max_len) + 's'
      print('\n******** Parsed arguments ********')
      for name, val in sorted(args.items()):
        print(max_len % name, ': ', val)
      print('**********************************\n')
    # convert it to struct
    return struct(args)

# ===========================================================================
# Simple math and processing
# ===========================================================================
def as_tuple(x, N=None, t=None):
  """
  Coerce a value to a tuple of given length (and possibly given type).

  Parameters
  ----------
  x : {value, iterable}
  N : {integer}
      length of the desired tuple
  t : {type, call-able, optional}
      required type for all elements

  Returns
  -------
  tuple
      ``tuple(x)`` if `x` is iterable, ``(x,) * N`` otherwise.

  Raises
  ------
  TypeError
      if `type` is given and `x` or any of its elements do not match it
  ValueError
      if `x` is iterable, but does not have exactly `N` elements

  Note
  ----
  This function is adpated from Lasagne
  Original work Copyright (c) 2014-2015 lasagne contributors
  All rights reserved.

  LICENSE: https://github.com/Lasagne/Lasagne/blob/master/LICENSE
  """
  # special case numpy array
  if not isinstance(x, tuple):
    if isinstance(x, (types.GeneratorType, list)):
      x = tuple(x)
    else:
      x = (x,)
  # ====== check length ====== #
  if is_number(N):
    N = int(N)
    if len(x) == 1:
      x = x * N
    elif len(x) != N:
      raise ValueError('x has length=%d, but required length N=%d' %
                       (len(x), N))
  # ====== check type ====== #
  if t is None:
    filter_func = lambda o: True
  elif isinstance(t, type) or isinstance(t, (tuple, list)):
    filter_func = lambda o: isinstance(o, t)
  elif hasattr(t, '__call__'):
    filter_func = t
  else:
    raise ValueError("Invalid value for `t`: %s" % str(t))
  if not all(filter_func(v) for v in x):
    raise TypeError("expected a single value or an iterable "
                    "of {0}, got {1} instead".format(t.__name__, x))
  return x

def as_tuple_of_shape(x):
  if not isinstance(x, (tuple, list)):
    x = (x,)
  if is_number(x[0]):
    x = (x,)
  return x

def as_list(x, N=None, t=None):
  return list(as_tuple(x, N, t))

def axis_normalize(axis, ndim,
                   return_tuple=False):
  """ Normalize the axis
   * -1 => ndim - 1
   * 10 => 10 % ndim
   * None => list(range(ndim))

  Parameters
  ----------
  return_tuple: bool
      if True, always return a tuple of normalized axis
  """
  if isinstance(axis, np.ndarray):
    axis = axis.ravel().tolist()
  if ndim == 0:
    return ()
  if axis is None or \
  (isinstance(axis, (tuple, list)) and len(axis) == 1 and axis[0] is None):
    return list(range(ndim))
  if not hasattr(axis, '__len__'):
    axis = int(axis) % ndim
    return (axis,) if return_tuple else axis
  axis = as_tuple(axis, t=int)
  return tuple([i % ndim for i in axis])


def flatten_list(x, level=None):
  """
  Parameters
  ----------
  level: int, or None
      how deep the function go into element of x to search for list and
      flatten it. If None is given, flatten all list found.

  Example
  -------
  >>> l = [1, 2, 3, [4], [[5], [6]], [[7], [[8], [9]]]]
  >>> print(flatten_list(l, level=1))
  >>> # [1, 2, 3, 4, [5], [6], [7], [[8], [9]]]
  >>> print(flatten_list(l, level=2))
  >>> # [1, 2, 3, 4, 5, 6, 7, [8], [9]]
  >>> print(flatten_list(l, level=None))
  >>> # [1, 2, 3, 4, 5, 6, 7, 8, 9]
  """
  if isinstance(x, Iterator):
    x = list(x)
  if level is None:
    level = 10e8
  if not isinstance(x, (tuple, list)):
    return [x]
  if any(isinstance(i, (tuple, list)) for i in x):
    _ = []
    for i in x:
      if isinstance(i, (tuple, list)) and level > 0:
        _ += flatten_list(i, level - 1)
      else:
        _.append(i)
    return _
  return x


# ===========================================================================
# Python
# ===========================================================================
class struct(dict):

  '''Flexible object can be assigned any attribtues'''

  def __init__(self, *args, **kwargs):
    super(struct, self).__init__(*args, **kwargs)
    # copy all dict to attr
    for i, j in self.items():
      if is_string(i) and not hasattr(self, i):
        super(struct, self).__setattr__(i, j)

  def __setattr__(self, name, val):
    super(struct, self).__setattr__(name, val)
    super(struct, self).__setitem__(name, val)

  def __setitem__(self, x, y):
    super(struct, self).__setitem__(x, y)
    if is_string(x):
      super(struct, self).__setattr__(x, y)


class bidict(dict):
  """ Bi-directional dictionary (i.e. a <-> b)
  Note
  ----
  When you iterate over this dictionary, it will be a doubled size
  dictionary
  """

  def __init__(self, *args, **kwargs):
    super(bidict, self).__init__(*args, **kwargs)
    # this is duplication
    self._inv = dict()
    for i, j in self.items():
      self._inv[j] = i

  @property
  def inv(self):
    return self._inv

  def __setitem__(self, key, value):
    super(bidict, self).__setitem__(key, value)
    self._inv[value] = key
    return None

  def __getitem__(self, key):
    if key not in self:
      return self._inv[key]
    return super(bidict, self).__getitem__(key)

  def update(self, *args, **kwargs):
    for k, v in dict(*args, **kwargs).items():
      self[k] = v
      self._inv[v] = k

  def __delitem__(self, key):
    del self._inv[super(bidict, self).__getitem__(key)]
    return dict.__delitem__(self, key)


# Under Python 2, 'urlretrieve' relies on FancyURLopener from legacy
# urllib module, known to have issues with proxy management
if sys.version_info[0] == 2:
  def urlretrieve(url, filename, reporthook=None, data=None):
    '''
    This function is adpated from: https://github.com/fchollet/keras
    Original work Copyright (c) 2014-2015 keras contributors
    '''
    def chunk_read(response, chunk_size=8192, reporthook=None):
      total_size = response.info().get('Content-Length').strip()
      total_size = int(total_size)
      count = 0
      while 1:
        chunk = response.read(chunk_size)
        if not chunk:
          break
        count += 1
        if reporthook:
          reporthook(count, chunk_size, total_size)
        yield chunk

    response = urlopen(url, data)
    with open(filename, 'wb') as fd:
      for chunk in chunk_read(response, reporthook=reporthook):
        fd.write(chunk)
else:
  from six.moves.urllib.request import urlretrieve

def get_file(fname, origin, outdir):
  '''
  Parameters
  ----------
  fname: output file name
  origin: url, link
  outdir: path to output dir
  '''
  fpath = os.path.join(outdir, fname)
  # ====== remove empty folder ====== #
  if os.path.exists(fpath):
    if os.path.isdir(fpath) and len(os.listdir(fpath)) == 0:
      shutil.rmtree(fpath)
  # ====== download package ====== #
  if not os.path.exists(fpath):
    prog = Progbar(target=-1,
                   name="Downloading: %s" % os.path.basename(origin),
                   print_report=True, print_summary=True)

    def dl_progress(count, block_size, total_size):
      if prog.target < 0:
        prog.target = total_size
      else:
        prog.add(count * block_size - prog.seen_so_far)
    error_msg = 'URL fetch failure on {}: {} -- {}'
    try:
      try:
        urlretrieve(origin, fpath, dl_progress)
      except URLError as e:
        raise Exception(error_msg.format(origin, e.errno, e.reason))
      except HTTPError as e:
        raise Exception(error_msg.format(origin, e.code, e.msg))
    except (Exception, KeyboardInterrupt) as e:
      if os.path.exists(fpath):
        os.remove(fpath)
      raise
  return fpath

# ===========================================================================
# Python utilities
# ===========================================================================
def get_all_files(path, filter_func=None):
  ''' Recurrsively get all files in the given path '''
  file_list = []
  if os.access(path, os.R_OK):
    for p in os.listdir(path):
      p = os.path.join(path, p)
      if os.path.isdir(p):
        file_list += get_all_files(p, filter_func)
      else:
        if filter_func is not None and not filter_func(p):
          continue
        # remove dump files of Mac
        if '.DS_Store' in p or '.DS_STORE' in p or \
            '._' == os.path.basename(p)[:2]:
          continue
        file_list.append(p)
  return file_list


def get_all_ext(path):
  """ Recurrsively get all extension of files in the given path

  Parameters
  ----------
  path : str
    input folder

  """
  file_list = []
  if os.access(path, os.R_OK):
    for p in os.listdir(path):
      p = os.path.join(path, p)
      if os.path.isdir(p):
        file_list += get_all_ext(p)
      else:
        # remove dump files of Mac
        if '.DS_Store' in p or '.DS_STORE' in p or \
            '._' == os.path.basename(p)[:2]:
          continue
        ext = p.split('.')
        if len(ext) > 1:
          file_list.append(ext[-1])
  file_list = list(set(file_list))
  return file_list


def folder2bin(path):
  """ This function read all files within a Folder
  in binary mode,
  then, store all the data in a dictionary mapping:
  `relative_path -> binary_data`
  """
  if not os.path.isdir(path):
    raise ValueError('`path`=%s must be a directory.' % path)
  path = os.path.abspath(path)
  files = get_all_files(path)
  data = {}
  for f in files:
    name = f.replace(path + '/', '')
    with open(f, 'rb') as f:
      data[name] = f.read()
  return data


def bin2folder(data, path, override=False):
  """ Convert serialized data from `folder2bin` back
  to a folder at `path`

  Parameters
  ----------
  data: {string, dict}
      if string, `data` can be pickled string, or path to a file.
      if dict, `data` is the output from `folder2bin`
  path: string
      path to a folder
  override: bool
      if True, override exist folder at `path`
  """
  # ====== check input ====== #
  if is_string(data):
    if os.path.isfile(data):
      with open(data, 'rb') as f:
        data = pickle.load(f)
    else:
      data = pickle.loads(data)
  if not isinstance(data, dict):
    raise ValueError("`data` must be dictionary type, or string, or path to file.")
  # ====== check outpath ====== #
  path = os.path.abspath(str(path))
  if not os.path.exists(path):
    os.mkdir(path)
  elif os.path.isfile(path):
    raise ValueError("`path` must be path to a directory.")
  elif os.path.isdir(path):
    if not override:
      raise RuntimeError("Folder at path:%s exist, cannot override." % path)
    shutil.rmtree(path)
    os.mkdir(path)
  # ====== deserialize ====== #
  for name, dat in data.items():
    with open(os.path.join(path, name), 'wb') as f:
      f.write(dat)
  return path


# ===========================================================================
# Package utils
# ===========================================================================
def package_installed(name, version=None):
  import pip
  for i in pip.get_installed_distributions():
    if name.lower() == i.key.lower() and \
    (version is None or version == i.version):
      return True
  return False


def package_list(include_version=False):
  """
  Return
  ------
  ['odin', 'lasagne', 'keras', ...] if include_version is False
  else ['odin==8.12', 'lasagne==25.18', ...]
  """

  all_packages = []
  import pip
  for i in pip.get_installed_distributions():
    all_packages.append(i.key +
        (('==' + i.version) if include_version is True else ''))
  return all_packages


def get_module_from_path(identifier, path='.',
                         prefix='', suffix='', exclude='',
                         prefer_compiled=False, return_error=False):
  ''' Algorithms:
   - Search all files in the `path` matched `prefix` and `suffix`
   - Exclude all files contain any str in `exclude`
   - Sorted all files based on alphabet
   - Load all modules based on `prefer_compiled`
   - return list of identifier found in all modules

  Parameters
  ----------
  identifier : str
      identifier of object, function or anything in script files

  prefix : str
      prefix of file to search in the `path`

  suffix : str
      suffix of file to search in the `path`

  path : str
      searching path of script files

  exclude : str, list(str)
      any files contain str in this list will be excluded

  prefer_compiled : bool
      True mean prefer .pyc file, otherwise prefer .py

  return_error : bool (default: False)
      return all the error happened during loading the modules

  Returns
  -------
  list(object, function, ..) :
      any thing match given identifier in all found script file
  errors (optional) :
      error while loading any of the script

  Notes
  -----
  File with multiple . character my procedure wrong results

  If the script run this this function match the searching process, a
  infinite loop may happen!

  * This function try to import each modules and find desire function,
  it may mess up something.

  '''
  import re
  import imp
  # ====== validate input ====== #
  if exclude == '':
    exclude = []
  if type(exclude) not in (list, tuple, numpy.ndarray):
    exclude = [exclude]
  prefer_flag = 1 if prefer_compiled else -1
  # ====== create pattern and load files ====== #
  pattern = re.compile(r"^%s.*%s\.pyc?" % (prefix, suffix)) # py or pyc
  file_name = os.listdir(path)
  file_name = [f for f in file_name
           if pattern.match(f) and
           sum([i in f for i in exclude]) == 0]
  # ====== remove duplicated pyc files ====== #
  file_name = sorted(file_name,
                     key=lambda x: prefer_flag * len(x)) # pyc is longer
  # .pyc go first get overrided by .py
  file_name = sorted({f.split('.')[0]: f for f in file_name}.values())
  # ====== load all modules ====== #
  modules = []
  modules_error = {}
  # NOTE: this will load the module ignore the relative import path
  # For example: for module A.B.C
  # `from A.B import C` will result 'A.B.C' -> C
  # `imp.load_source`, here, will result 'C' -> C
  for fname in file_name:
    try:
      if '.pyc' in fname:
        modules.append(
            imp.load_compiled(fname.split('.')[0],
                              os.path.join(path, fname))
        )
      else:
        modules.append(
            imp.load_source(fname.split('.')[0],
                            os.path.join(path, fname))
        )
    except Exception as e:
      modules_error[fname] = str(e)
      eprint("Cannot loading modules from file: '%s' - %s" %
        (ctext(fname, 'yellow'), ctext(str(e), 'red')))
  # ====== Find all identifier in modules ====== #
  ids = []
  for m in modules:
    for i in inspect.getmembers(m):
      if identifier in i:
        ids.append(i[1])
  # remove duplicate py
  if bool(return_error):
    return ids, modules_error
  return ids


def ordered_set(seq):
  seen = {}
  result = []
  for marker in seq:
    if marker in seen: continue
    seen[marker] = 1
    result.append(marker)
  return result


def dict_union(*dicts, **kwargs):
  r"""Return union of a sequence of disjoint dictionaries.

  Parameters
  ----------
  dicts : dicts
      A set of dictionaries with no keys in common. If the first
      dictionary in the sequence is an instance of `OrderedDict`, the
      result will be OrderedDict.
  \*\*kwargs
      Keywords and values to add to the resulting dictionary.

  Raises
  ------
  ValueError
      If a key appears twice in the dictionaries or keyword arguments.

  """
  dicts = list(dicts)
  if dicts and isinstance(dicts[0], OrderedDict):
    result = OrderedDict()
  else:
    result = {}
  for d in list(dicts) + [kwargs]:
    duplicate_keys = set(result.keys()) & set(d.keys())
    if duplicate_keys:
      raise ValueError("The following keys have duplicate entries: {}"
                       .format(", ".join(str(key) for key in
                                         duplicate_keys)))
    result.update(d)
  return result


# ===========================================================================
# PATH, path manager
# ===========================================================================
@contextlib.contextmanager
def TemporaryDirectory(add_to_path=False):
  """
  add_to_path: bool
      temporary add the directory to system $PATH
  """
  temp_dir = tempfile.mkdtemp()
  if add_to_path:
    os.environ['PATH'] = temp_dir + ':' + os.environ['PATH']
  current_dir = os.getcwd()
  os.chdir(temp_dir)
  yield temp_dir
  os.chdir(current_dir)
  if add_to_path:
    os.environ['PATH'] = os.environ['PATH'].replace(temp_dir + ':', '')
  shutil.rmtree(temp_dir)


def get_tempdir():
  return tempfile.mkdtemp()


def _get_managed_path(folder, name, override, is_folder=False,
                      root='~', odin_base=True):
  if root == '~':
    root = os.path.expanduser('~')
  datadir_base = os.path.join(root, '.odin') if odin_base else root
  if not os.path.exists(datadir_base):
    os.mkdir(datadir_base)
  elif not os.access(datadir_base, os.W_OK):
    raise Exception('Cannot acesss path: ' + datadir_base)
  datadir = os.path.join(datadir_base, folder)
  if not os.path.exists(datadir):
    os.makedirs(datadir)
  # ====== check given path with name ====== #
  if is_string(name):
    datadir = os.path.join(datadir, str(name))
    if os.path.exists(datadir) and override:
      if os.path.isfile(datadir): # remove file
        os.remove(datadir)
      else: # remove and create new folder
        shutil.rmtree(datadir)
    if is_folder and not os.path.exists(datadir):
      os.mkdir(datadir)
  return datadir


def get_datasetpath(name=None, override=False, is_folder=True, root='~'):
  return _get_managed_path('datasets', name, override,
                           is_folder=is_folder, root=root)


def get_figpath(name=None, override=False, root='~'):
  return _get_managed_path('figs', name, override,
                           is_folder=True, root=root)

def get_modelpath(name=None, override=False, root='~'):
  """ Default model path for saving O.D.I.N networks """
  return _get_managed_path('models', name, override,
                           is_folder=True, root=root)


def get_logpath(name=None, override=False, increasing=True,
                odin_base=True, root='~'):
  """
  name : (string) name of log file
  override : (bool) if True, override old log file if exist
  increasing : (bool) append a number (e.g. "log.0.txt"...), if log file exists
  odin_base : (bool) if True, create log file under '.odin/logs' folder
  root : (string) root folder that contain all the log file
  """
  def _check_logpath(log_path):
    main_path, ext = os.path.splitext(log_path)
    main_path = main_path.split('.')
    try:
      current_log_index = int(main_path[-1])
      main_path = main_path[:-1]
    except ValueError:
      current_log_index = 0
    main_path = '.'.join(main_path)
    # ====== increase log index until found a new file ====== #
    while True:
      path = main_path + '.' + str(current_log_index) + ext
      if not os.path.exists(path):
        break
      current_log_index += 1
    return main_path + '.' + str(current_log_index) + ext

  path = _get_managed_path('logs' if odin_base else '', name, override,
                           is_folder=False, root=root,
                           odin_base=odin_base)
  if increasing:
    path = _check_logpath(path)
  return path


def get_exppath(tag, name=None, override=False, prompt=False,
                root='~'):
  """ Specific path for experiments results

  Parameters
  ----------
  tag: string
      specific tag for the task you are working on
  name: string
      name of the folder contains all the results (NOTE: the
      name can has subfolder)
  override: bool
      if True, remove exist folder
  prompt: bool
      if True, display prompt and require (Y) input before
      delete old folder, if (N), the program exit.
  root: string
      root path for the results (default: "~/.odin")
  """
  path = _get_managed_path('exp', tag, False,
                           is_folder=True, root=root,
                           odin_base=False)
  # only return the main folder
  if name is None:
    pass
  # return the sub-folder
  else:
    name = str(name).split('/')
    for i in name:
      path = os.path.join(path, i)
      if not os.path.exists(path):
        os.mkdir(path)
  # ====== check if override ====== #
  if override and len(os.listdir(path)) > 0:
    if prompt:
      user_cmd = raw_input('Do you want to delete "%s" (Y for yes):').lower()
      if user_cmd != 'y':
        exit()
    shutil.rmtree(path)
    os.mkdir(path)
  return path

# ===========================================================================
# Misc
# ===========================================================================
def run_script(s, wait=True, path='/tmp'):
  """
  Parameters
  ----------
  s : string
    python script

  wait : bool (default: True)
    blocking the current process until finishing the script

  path : string (default: '/tmp')
    saving the script to a temporary path before running

  Return
  ------
  status: of executed command, if `wait`=True else return Popen object
  out: (string, utf-8) - stdout message
  err: (string, utf-8) - stderr message
  """
  # ====== path preprocessing ====== #
  if path is None:
    path = '/tmp'
  if os.path.isdir(path):
    path = os.path.join(path, 'tmp_%s' % uuid(length=25) + '.py')
  path = os.path.abspath(path)
  # ====== script preprocessing ====== #
  s = str(s).strip()
  # ====== run the script ====== #
  try:
    with open(path, 'w') as f:
      f.write(s)
    cmd = subprocess.Popen("python %s" % path, shell=True,
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (out, err) = cmd.communicate()
    if wait:
      status = cmd.wait()
    else:
      status = cmd
  finally:
    if os.path.exists(path):
      os.remove(path)
  return status, str(out, encoding='utf-8'), str(err, encoding='utf-8')

def exec_commands(cmds, print_progress=True):
  ''' Execute a command or list of commands in parallel with multiple process
  (as much as we have CPU)

  Parameters
  ----------
  cmds: str or list of str
      string represent command you want to run

  Return
  ------
  failed: list of failed command

  '''
  if not cmds: return [] # empty list
  if not isinstance(cmds, (list, tuple)):
    cmds = [cmds]
  else:
    cmds = list(cmds)

  def done(p):
    return p.poll() is not None

  def success(p):
    return p.returncode == 0

  max_task = cpu_count()
  processes = []
  processes_map = {}
  failed = []
  if print_progress:
    prog = Progbar(target=len(cmds), name="Execute Commands")
  while True:
    while cmds and len(processes) < max_task:
      task = cmds.pop()
      p = subprocess.Popen(task, shell=True)
      processes.append(p)
      processes_map[p] = task
      # print process
      if print_progress:
        prog.add(1)

    for p in processes:
      if done(p):
        if success(p):
          processes.remove(p)
        else:
          failed.append(processes_map[p])

    if not processes and not cmds:
      break
    else:
      time.sleep(0.005)
  return failed


def save_wav(path, s, fs):
  from scipy.io import wavfile
  wavfile.write(path, fs, s)


def play_audio(data, fs, volumn=1, speed=1):
  ''' Play audio from numpy array.

  Parameters
  ----------
  data : numpy.ndarray
          signal data
  fs : int
          sample rate
  volumn: float
          between 0 and 1
  speed: float
          > 1 mean faster, < 1 mean slower

  Note
  ----
  Only support play audio on MacOS
  '''
  import soundfile as sf
  import os

  data = numpy.asarray(data, dtype=numpy.int16)
  if data.ndim == 1:
    channels = 1
  else:
    channels = data.shape[1]
  with TemporaryDirectory() as temppath:
    path = os.path.join(temppath, 'tmp_play.wav')
    with sf.SoundFile(path, 'w', fs, channels, subtype=None,
        endian=None, format=None, closefd=None) as f:
      f.write(data)
    os.system('afplay -v %f -q 1 -r %f %s &' % (volumn, speed, path))
    raw_input('<enter> to stop audio.')
    os.system("kill -9 `ps aux | grep -v 'grep' | grep afplay | awk '{print $2}'`")


# ===========================================================================
# System query
# ===========================================================================
__process_pid_map = {}


def get_process_status(pid=None, memory_usage=False, memory_shared=False,
                       memory_virtual=False, memory_maps=False,
                       cpu_percent=False, threads=False,
                       status=False, name=False, io_counters=False):
  import psutil
  if pid is None:
    pid = os.getpid()
  if pid in __process_pid_map:
    process = __process_pid_map[pid]
  else:
    process = psutil.Process(pid)
    __process_pid_map[pid] = process

  if status:
    return process.status()
  if name:
    return process.name()
  if io_counters:
    return process.io_counters()
  if memory_usage:
    return process.memory_info().rss / float(2 ** 20)
  if memory_shared:
    return process.memory_info().shared / float(2 ** 20)
  if memory_virtual:
    return process.memory_info().vms / float(2 ** 20)
  if memory_maps:
    return {i[0]: i[1] / float(2 ** 20)
            for i in process.memory_maps()}
  if cpu_percent:
    # first time call always return 0
    process.cpu_percent(None)
    return process.cpu_percent(None)
  if threads:
    return {i.id: (i.user_time, i.system_time) for i in process.threads()}


def get_system_status(memory_total=False, memory_total_actual=False,
                      memory_total_usage=False, memory_total_free=False,
                      all_pids=False, swap_memory=False, pid=False):
  """
  Parameters
  ----------
  threads: bool
      return dict {id: (user_time, system_time)}
  memory_maps: bool
      return dict {path: rss}

  Note
  ----
  All memory is returned in `MiB`
  To calculate memory_percent:
      get_system_status(memory_usage=True) / get_system_status(memory_total=True) * 100
  """
  import psutil
  # ====== general system query ====== #
  if memory_total:
    return psutil.virtual_memory().total / float(2 ** 20)
  if memory_total_actual:
    return psutil.virtual_memory().available / float(2 ** 20)
  if memory_total_usage:
    return psutil.virtual_memory().used / float(2 ** 20)
  if memory_total_free:
    return psutil.virtual_memory().free / float(2 ** 20)
  if swap_memory:
    tmp = psutil.swap_memory()
    tmp.total /= float(2 ** 20)
    tmp.used /= float(2 ** 20)
    tmp.free /= float(2 ** 20)
    tmp.sin /= float(2**20)
    tmp.sout /= float(2**20)
    return tmp
  if all_pids:
    return psutil.pids()
  if pid:
    return os.getpid()



from __future__ import print_function, division, absolute_import

import tensorflow as tf
from tensorflow.python.ops import init_ops

from odin import backend as K
from odin.autoconfig import randint
from odin.nnet.base import NNOp


class Embedding(NNOp):
  """ Embedding
  Parameters
  ----------
  input_size: int
      size of dictionary

  output_size : int
      number of dimension of the vector to represent each entity
      in the dictionary.

  W_inti : trainable variable, expression, numpy array or call-able
      Initial value (or specification) for initialize weights matrix
      which has size (input_szie, output_size)
  """

  def __init__(self, input_size, output_size,
               W_init=init_ops.random_uniform_initializer(seed=randint()), **kwargs):
    super(Embedding, self).__init__(**kwargs)
    self.input_size = input_size
    self.output_size = output_size
    self.W_init = W_init

  @property
  def embedding_shape(self):
    """ Return the estimated embedding matrix shape """
    return (self.input_size, self.output_size)

  def _initialize(self):
    self.get_variable_nnop(initializer=self.W_init,
                           shape=(self.input_size, self.output_size),
                           name='W', roles=K.role.EmbeddingWeight)

  def _apply(self, X):
    return tf.gather(self.W, tf.cast(X, tf.int32))



# -*- coding: utf-8 -*-
from __future__ import division, absolute_import, print_function

import os
import re
import sys
import time
import inspect
from enum import Enum
from numbers import Number
from datetime import datetime
from collections import defaultdict
from abc import ABCMeta, abstractproperty, abstractmethod
from six import add_metaclass, string_types
from six.moves import cPickle

import numpy as np
import scipy as sp

from odin.utils import (as_tuple, is_string, add_notification, ctext,
                        FuncDesc)

__all__ = [
    'TrainSignal',
    'TaskSignal',
    'Callback',
    'Debug',
    'LambdaCallback',
    'CallbackList',
    'NaNDetector',
    'CheckpointEpoch',
    'CheckpointGeneralization',
    'EarlyStop',
    'EarlyStopGeneralizationLoss',
    'EarlyStopPatience',
    'LRdecay',
    'EpochSummary'
]

# This SIGNAL can terminate running iterator (or generator),
# however, it cannot terminate them at the beginning,
# so you should not call .send(SIG_TERMINATE_ITERATOR)
# at beginning of iterator
class TrainSignal(Enum):
  """ TrainingSignal """
  NONE = 0 # nothing happened
  SAVE = 1 # save the parameters during training
  SAVE_BEST = 2 # saving the best model, signaled by EarlyStop
  ROLLBACK = 3 # rollback the model to the best checkpoint
  STOP = 4 # stop the training task

class TaskSignal(Enum):
  """ Signal represent current activities of a Training, Validating or
  Evaluating task """
  BatchStart = 0
  BatchEnd = 1
  EpochStart = 2
  EpochEnd = 3
  TaskStart = 4
  TaskEnd = 5

# ===========================================================================
# Helpers
# ===========================================================================
def _parse_result(result):
  if isinstance(result, (tuple, list)) and len(str(result)) > 20:
    type_str = ''
    if len(result) > 0:
      type_str = type(result[0]).__name__
    return 'list;%d;%s' % (len(result), type_str)
  s = str(result)
  return s[:20]

TASK_TYPES = ['task', 'subtask', 'crosstask', 'othertask']

def time2date(timestamp):
  return datetime.fromtimestamp(timestamp).strftime('%y-%m-%d %H:%M:%S')

def date2time(date):
  return time.mktime(datetime.datetime.strptime(date, '%y-%m-%d %H:%M:%S').timetuple())
# ===========================================================================
# Callbacks
# ===========================================================================
class Callback(object):

  """Callback
  Order of execution:
   - task_start(self, task)
   - epoch_start(self, task, data)
   - batch_start(self, task, data)
   - batch_end(self, task, results)
   - epoch_end(self, task, results)
   - task_end(self, task, results)

  For `batch_start` and `epoch_start`, `data` is the input data to training, validating
  or any Task

  For `batch_end`, `results` is the mapping `tensor_name` -> `returned_value`

  For `epoch_end`, `results` is the mapping `tensor_name` -> `list_of_returned_value`

  For `task_end`, results is the mapping `epoch_index` -> defaultdict(`epoch_results`),
  where `epoch_results` is the results of `epoch_end`.

  `task` is an instance of `odin.training.trainer.Task`, which
  indicate which Task is running (e.g. Training, Validating, ...)
  Some accessible properties from `odin.training.Task`:
   - curr_epoch: Total number of epoch finished since the beginning of the Task
   - curr_iter: Total number of iteration finished since the beginning of the Task
   - curr_samples: Total number of samples finished since the beginning of the Task
   - curr_epoch_iter: Number of iteration within current epoch
   - curr_epoch_samples: Number of samples within current epoch
  """

  def __init__(self, logging=True):
    super(Callback, self).__init__()
    self._log = bool(logging)

  def set_notification(self, is_enable):
    """ Turn notification on and off """
    self._log = bool(is_enable)
    return self

  def batch_start(self, task, batch):
    pass

  def batch_end(self, task, batch_results):
    pass

  def epoch_start(self, task, data):
    pass

  def epoch_end(self, task, epoch_results):
    pass

  def task_start(self, task):
    pass

  def task_end(self, task, task_results):
    pass

  def event(self, event_name):
    """ This function is directly called by MainLoop when
    special event triggered, `odin.training.callback.TrainSignal`
    """
    pass

  def send_notification(self, msg):
    if self._log:
      add_notification(
          '[%s] %s' % (ctext(self.__class__.__name__, 'magenta'), msg))
    return self

class CallbackList(Callback):

  ''' Broadcast signal to all its children'''

  def __init__(self, callbacks=None):
    super(CallbackList, self).__init__()
    self.set_callbacks(callbacks)

  def set_callbacks(self, callbacks):
    if callbacks is None:
      callbacks = []
    elif isinstance(callbacks, CallbackList):
      callbacks = callbacks._callbacks
    else:
      callbacks = as_tuple(callbacks,
        t=lambda x: isinstance(x, (Callback, type(None))))
      callbacks = [i for i in callbacks if i is not None]
    self._callbacks = [i for i in set(callbacks)]
    return self

  def set_notification(self, is_enable):
    """ Turn notification on and off """
    is_enable = bool(is_enable)
    self._log = is_enable
    for cb in self._callbacks:
      cb.set_notification(is_enable)
    return self

  def __str__(self):
    return '<CallbackList: ' + \
    ', '.join([i.__class__.__name__ for i in self._callbacks]) + '>'

  def batch_start(self, task, batch):
    msg = []
    for i in self._callbacks:
      m = i.batch_start(task, batch)
      msg += [j for j in as_tuple(m) if isinstance(j, TrainSignal)]
    return msg

  def batch_end(self, task, batch_results):
    msg = []
    for i in self._callbacks:
      m = i.batch_end(task, batch_results)
      msg += [j for j in as_tuple(m) if isinstance(j, TrainSignal)]
    return msg

  def epoch_start(self, task, data):
    msg = []
    for i in self._callbacks:
      m = i.epoch_start(task, data)
      msg += [j for j in as_tuple(m) if isinstance(j, TrainSignal)]
    return msg

  def epoch_end(self, task, epoch_results):
    msg = []
    for i in self._callbacks:
      m = i.epoch_end(task, epoch_results)
      msg += [j for j in as_tuple(m) if isinstance(j, TrainSignal)]
    return msg

  def task_start(self, task):
    msg = []
    for i in self._callbacks:
      m = i.task_start(task)
      msg += [j for j in as_tuple(m) if isinstance(j, TrainSignal)]
    return msg

  def task_end(self, task, task_results):
    msg = []
    for i in self._callbacks:
      m = i.task_end(task, task_results)
      msg += [j for j in as_tuple(m) if isinstance(j, TrainSignal)]
    return msg

  def event(self, event_type):
    assert isinstance(event_type, TrainSignal), \
    "event must be enum `TrainSignal` but given: %s" % str(type(event_type))
    for i in self._callbacks:
      i.event(event_type)

class Debug(Callback):
  """ Debug """

  def __init__(self, signal=None):
    super(Debug, self).__init__()
    if signal is None:
      pass

  def batch_start(self, task, batch):
    print("Batch Start:", task.name, task.curr_epoch, task.curr_samples,
          [(i.shape, i.dtype, type(i)) for i in batch])

  def batch_end(self, task, batch_results):
    print(ctext("Batch End:", 'cyan'),
          task.name, task.curr_epoch, task.curr_samples,
          [(i.shape, i.dtype, type(i)) for i in as_tuple(batch_results)])

  def epoch_start(self, task, data):
    print(ctext("Epoch Start:", 'cyan'),
          task.name, task.curr_epoch, task.curr_samples,
          [(i.shape, i.dtype, type(i)) for i in data])

  def epoch_end(self, task, epoch_results):
    print(ctext("Epoch End:", 'cyan'),
          task.name, task.curr_epoch, task.curr_samples,
          [(i, len(j), type(j[0])) for i, j in epoch_results.items()])

  def task_start(self, task):
    print(ctext("Task Start:", 'cyan'),
          task.name, task.curr_epoch, task.curr_samples)

  def task_end(self, task, task_results):
    print(ctext("Task End:", 'cyan'),
        task.name, task.curr_epoch, task.curr_samples,
        [(i, [(n, len(v), type(v[0])) for n, v in j.items()])
         for i, j in task_results.items()])

  def event(self, event_type):
    print(ctext("[Debug] Event:", 'cyan'), event_type)

# ===========================================================================
# Others
# ===========================================================================
class LambdaCallback(Callback):
  """ Some accessible properties from `odin.training.Task`:
   - curr_epoch: Total number of epoch finished since the beginning of the Task
   - curr_iter: Total number of iteration finished since the beginning of the Task
   - curr_samples: Total number of samples finished since the beginning of the Task
   - curr_epoch_iter: Number of iteration within current epoch
   - curr_epoch_samples: Number of samples within current epoch

  Parameters
  ----------
  fn : function
    a function, must follow one of the option:
      - 0 argument: only the function is called
      - 1 argument: only instance of `odin.training.Task`
      - 2 arguments: instance of `odin.training.Task`, and
                     input_data or returned_results
  name : string
    name of the Task when this callback will be activated
  signal : odin.training.callbacks.TaskSignal
    signal that trigger this callback

  """

  def __init__(self, fn, task_name,
               signal=TaskSignal.EpochEnd):
    super(LambdaCallback, self).__init__()
    assert hasattr(fn, '__call__')
    # ====== check function ====== #
    spec = inspect.signature(fn)
    assert len(spec.parameters) in (0, 1, 2),\
    "`fn` must accept 0, 1 or 2 arguments"
    self.n_args = len(spec.parameters)
    self.fn = FuncDesc(fn)
    # ====== others ====== #
    self.task_name = str(task_name)
    assert isinstance(signal, TaskSignal)
    self.signal = signal

  def _call_fn(self, task, data):
    if self.n_args == 0:
      return self.fn()
    elif self.n_args == 1:
      return self.fn(task)
    elif self.n_args == 2:
      return self.fn(task, data)
    else:
      raise RuntimeError()

  def batch_start(self, task, data):
    if self.signal == TaskSignal.BatchStart and \
    self.task_name == task.name:
      self._call_fn(task, data)

  def batch_end(self, task, batch_results):
    if self.signal == TaskSignal.BatchEnd and \
    self.task_name == task.name:
      self._call_fn(task, batch_results)

  def epoch_start(self, task, data):
    if self.signal == TaskSignal.EpochStart and \
    self.task_name == task.name:
      self._call_fn(task, data)

  def epoch_end(self, task, epoch_results):
    if self.signal == TaskSignal.EpochEnd and \
    self.task_name == task.name:
      self._call_fn(task, epoch_results)

  def task_start(self, task):
    if self.signal == TaskSignal.TaskStart and \
    self.task_name == task.name:
      self._call_fn(task, None)

  def task_end(self, task, task_results):
    if self.signal == TaskSignal.TaskEnd and \
    self.task_name == task.name:
      self._call_fn(task, task_results)

# ===========================================================================
# NaN value detection
# ===========================================================================
class NaNDetector(Callback):
  """ NaNDetector

  Parameters
  ----------
  task_name : {str, None}
    name of specific Task will be applied,
    if `None`, all tasks are considered

  patience : {int}
    the Task will be stopped if `patience` < 0, `patience` will
    subtracted by 1 every time NaN is detected

  logging : bool (default: True)
    show notification when NaN detected

  """

  def __init__(self, task_name=None, patience=-1,
               detect_inf=False, logging=True):
    super(NaNDetector, self).__init__(logging=logging)
    self._task_name = task_name
    self._patience = patience
    self._detect_inf = bool(detect_inf)

  def batch_end(self, task, batch_results):
    if self._task_name is not None and task.name != self._task_name:
      return
    # found any NaN values
    if self._detect_inf:
      fn = lambda x: np.logical_or(np.isinf(x), np.isnan(x))
    else:
      fn = lambda x: np.isnan(x)

    if any(np.any(fn(r)) for r in as_tuple(batch_results)):
      signal = TrainSignal.ROLLBACK
      self._patience -= 1
      if self._patience <= 0: # but if out of patience, stop
        signal = TrainSignal.STOP
      self.send_notification('Found NaN or Inf value, task:"%s"' % task.name)
      return signal

# ===========================================================================
# Checkpoint technique
# ===========================================================================
class CheckpointEpoch(Callback):
  """ CheckpointEpoch

  Periodically saving the model after fixed percents of epoch have
  passed (using the estimated number of iteration to calculate current
  epoch percents)

  """

  def __init__(self, task_name, epoch_percent=1., logging=True):
    super(CheckpointEpoch, self).__init__(logging=logging)
    self._task_name = task_name

    epoch_percent = float(epoch_percent)
    self._epoch_percent = epoch_percent

  def batch_end(self, task, batch_results):
    if task.name == self._task_name:
      if task.curr_epoch_iter % int(self._epoch_percent * task.iter_per_epoch) == 0:
        return TrainSignal.SAVE
    return None

class CheckpointGeneralization(Callback):
  """ Save the model when improvement on generalization score

  Parameters
  ----------
  task_name : string
      task name for checking this criterion

  output_name : {Tensor, string}
      name of the Tensor or the Tensor itself for monitoring

  improvement_margin : float (>= 0)
      improvement in percentage that will be accepted,
      default, 0% mean that any number is smaller will trigger
      the checkpoint.
  """

  def __init__(self, task_name, output_name,
               lower_better=True, improvement_margin=0, logging=True):
    super(CheckpointGeneralization, self).__init__(logging=logging)
    self._task_name = str(task_name)
    self._output_name = output_name if is_string(output_name) \
        else output_name.name
    self._lower_better = bool(lower_better)
    self._improvement_margin = float(improvement_margin)
    assert self._improvement_margin >= 0
    self._best_score = None

  def epoch_end(self, task, epoch_results):
    if task.name != self._task_name:
      return None
    if self._output_name not in epoch_results:
      return None
    # get the score
    score = np.mean(epoch_results[self._output_name])
    if not self._lower_better:
      score *= -1
    epsilon = 1e-8
    # first time, nothing to do
    if self._best_score is None:
      self._best_score = score
      return None

    gl_t = score / (self._best_score + epsilon)
    if gl_t < 1 - (self._improvement_margin / 100):
      self._best_score = score
      return TrainSignal.SAVE_BEST

# ===========================================================================
# Learning rate manipulation
# ===========================================================================
class LRdecay(Callback):
  """ LRdecay
  whenever TrainSignal.ROLLBACK is triggered, decrease the learning
  rate by `decay_rate`
  """

  def __init__(self, lr, decay_rate=0.5):
    super(LRdecay, self).__init__()
    from odin import backend as K
    self.lr = lr
    self.lr_value = K.get_value(lr)
    self.decay_rate = decay_rate

  def event(self, event_name):
    if event_name == TrainSignal.ROLLBACK:
      from odin import backend as K
      self.lr_value *= self.decay_rate
      K.set_value(self.lr, self.lr_value)

# ===========================================================================
# EarlyStop utilities
# ===========================================================================
@add_metaclass(ABCMeta)
class EarlyStop(Callback):
  """
  Early Stopping algorithm based on Generalization Loss criterion,
  this is strict measure on validation

  ``LOWER is better``

  Parameters
  ----------
  task_name : string
      task name for checking this criterion

  output_name : {Tensor, string}
      name of the Tensor or the Tensor itself for monitoring

  threshold : float
      for example, threshold = 5, if we loss 5% of performance on validation
      set, then stop

  patience: int
      how many cross the threshold that still can be rollbacked

  get_value : function
      function to process the results of whole epoch (i.e list of results
      returned from batch_end) to return comparable number.
      For example, lambda x: np.mean(x)

  Note
  ----
  * The early stop checking will be performed at the end of an epoch.
  * By default, the return value from epoch mean the loss value, i.e lower
  is better
  * If multiple value returned, you have to modify the get_value function
  """

  def __init__(self, task_name, output_name, threshold, patience=1,
               get_value=lambda x: np.mean(x),
               logging=True):
    super(EarlyStop, self).__init__(logging=logging)
    self._task_name = str(task_name)
    self._output_name = output_name if is_string(output_name) \
        else output_name.name

    self._threshold = float(threshold)
    self._patience = int(patience)

    if get_value is None:
      get_value = lambda x: x
    elif not hasattr(get_value, '__call__'):
      raise ValueError('get_value must call-able')
    self._get_value = get_value
    # ====== history ====== #
    self._history = []

  # ==================== main callback methods ==================== #
  def epoch_end(self, task, epoch_results):
    if self._task_name != task.name:
      return
    self._history.append(self._get_value(epoch_results[self._output_name]))
    # ====== check early stop ====== #
    shouldSave, shouldStop = self.earlystop(self._history, self._threshold)
    msg = None
    if shouldSave > 0:
      msg = TrainSignal.SAVE_BEST
    if shouldStop > 0:
      msg = TrainSignal.ROLLBACK
      # check patience
      self._patience -= 1
      if self._patience < 0:
        msg = TrainSignal.STOP
    self.send_notification('Message "%s"' % str(msg))
    return msg

  @abstractmethod
  def earlystop(self, history, threshold):
    """ Any algorithm return: shouldSave, shouldStop """
    pass

class EarlyStopGeneralizationLoss(EarlyStop):
  """ Early Stopping algorithm based on Generalization Loss criterion,
  this is strict measure on validation

  ``LOWER is better``

  Parameters
  ----------
  name : string
      task name for checking this criterion
  threshold : float
      for example, threshold = 5, if we loss 5% of performance on validation
      set, then stop
  patience: int
      how many cross the threshold that still can be rollbacked
  get_value : function
      function to process the results of whole epoch (i.e list of results
      returned from batch_end) to return comparable number.

  Note
  ----
  The early stop checking will be performed at the end of an epoch.
  By default, the return value from epoch mean the loss value, i.e lower
  is better.
  By default, the `get_value` function will only take the first returned
  value for evaluation.
  """

  def __init__(self, task_name, output_name, threshold=5, patience=1,
               get_value=lambda x: np.mean(x), logging=True):
    super(EarlyStopGeneralizationLoss, self).__init__(
        task_name, output_name, threshold, patience,
        get_value, logging=logging)

  def earlystop(self, history, threshold):
    gl_exit_threshold = threshold
    longest_remain_performance = int(gl_exit_threshold + 1)
    epsilon = 1e-8

    if len(history) == 0: # no save, no stop
      return 0, 0
    shouldStop = 0
    shouldSave = 0

    gl_t = 100 * (history[-1] / (min(history) + epsilon) - 1)
    if gl_t <= 0 and np.argmin(history) == (len(history) - 1):
      shouldSave = 1
      shouldStop = -1
    elif gl_t > gl_exit_threshold:
      shouldStop = 1
      shouldSave = -1

    # check stay the same performance for so long
    if len(history) > longest_remain_performance:
      remain_detected = 0
      j = history[-longest_remain_performance]
      for i in history[-longest_remain_performance:]:
        if abs(i - j) < 1e-5:
          remain_detected += 1
      if remain_detected >= longest_remain_performance:
        shouldStop = 1
    return shouldSave, shouldStop

class EarlyStopPatience(EarlyStop):
  """
  EarlyStopPatience(self, name, threshold, patience=1,
            get_value=lambda x: np.mean([i[0] for i in x]
                                        if isinstance(x[0], (tuple, list))
                                        else x),
            stop_callback=None,
            save_callback=None)


  Adapted algorithm from keras:
  All contributions by François Chollet:
  Copyright (c) 2015, François Chollet.
  All rights reserved.

  All contributions by Google:
  Copyright (c) 2015, Google, Inc.
  All rights reserved.

  All other contributions:
  Copyright (c) 2015, the respective contributors.
  All rights reserved.

  LICENSE: https://github.com/fchollet/keras/blob/master/LICENSE

  Stop training when a monitored quantity has stopped improving.

  Parameters
  ----------
  name : string
      task name for checking this criterion
  threshold : float
      for example, threshold = 5, if we loss 5% of performance on validation
      set, then stop
  patience: int
      how many cross the threshold that still can be rollbacked
  get_value : function
      function to process the results of whole epoch (i.e list of results
      returned from batch_end) to return comparable number.
      for example, lambda x: np.mean(x)

  """

  def __init__(self, task_name, output_name, threshold, patience=1,
               get_value=lambda x: np.mean(x), logging=True):
    super(EarlyStopPatience, self).__init__(
        task_name, output_name, threshold, patience,
        get_value, logging=logging)

  def earlystop(self, history, threshold):
    if not hasattr(self, 'wait'): self.wait = 0
    shouldSave, shouldStop = 0, 0
    # showed improvement, should not equal to old best
    if len(history) <= 1 or history[-1] < np.min(history[:-1]):
      self.wait = 0
      shouldSave = 1
    else:
      if self.wait >= threshold:
        shouldSave = -1
        shouldStop = 1
      self.wait += 1
    return shouldSave, shouldStop

# ===========================================================================
# Monitoring
# ===========================================================================
class EpochSummary(Callback):
  """
  Parameters
  ----------
  output_name : {list of string, list of Tensor}
    list of outputs for monitoring

  print_plot : bool (default: False)
    using bashplot to print out the figure directly in terminal

  save_path : {str, None} (default: None)
    if provided, save pdf figure to given path

  epoch_percent : float (0 < .. < 1)
    determine the periodic of the summary within single epoch

  repeat_freq : int (.. > 0)
    determine the repeating frequency (e.g. `repeat_freq=2`
    is for every 2 epoch)

  fn_reduce : bool (default:False)


  """

  def __init__(self, task_name, output_name,
               fn_reduce=lambda x: (np.mean(x)
                                    if isinstance(x[0], Number) else
                                    sum(i for i in x)),
               print_plot=False, save_path=None,
               repeat_freq=1, logging=True):
    super(EpochSummary, self).__init__(logging=logging)
    self._task_name = as_tuple(task_name, t=str)
    # ====== scheduling ====== #
    assert repeat_freq >= 1
    self._repeat_freq = int(repeat_freq)
    self._count = self._repeat_freq * len(self._task_name)
    self._epoch_results = defaultdict(dict)
    # ====== output identity ====== #
    if not isinstance(output_name, (tuple, list, set)):
      output_name = (output_name,)
    output_name = [i if is_string(i) else i.name
                   for i in output_name]
    self.output_name = tuple(output_name)
    self.fn_reduce = FuncDesc(func=fn_reduce)
    # ====== how to output ====== #
    self.print_plot = bool(print_plot)
    self.save_path = save_path

  def epoch_end(self, task, epoch_results):
    output_name = self.output_name
    if len(output_name) == 0: # nothing to do
      return
    task_name = self._task_name

    if task.name in task_name:
      self._count -= 1
      # ====== processing results ====== #
      assert all(name in epoch_results for name in output_name),\
      "Given outputs with name: %s; but task: '%s' results only contain name: %s" % \
      (', '.join(self.output_name), str(task), ', '.join(tuple(epoch_results.keys())))

      for name in output_name:
        batch_results = epoch_results[name]
        if name not in self._epoch_results[task.name]:
          self._epoch_results[task.name][name] = []
        self._epoch_results[task.name][name].append(self.fn_reduce(batch_results))
      # ====== start plotting ====== #
      if self._count == 0:
        self._count = self._repeat_freq * len(task_name)
        from odin import visual as V
        n_col = len(task_name)
        n_row = len(output_name)
        if self.save_path is not None:
          from matplotlib import pyplot as plt
        save_figures = False
        override = True

        for o_idx, o_name in enumerate(output_name):
          results = {task_name: r[o_name]
                     for task_name, r in self._epoch_results.items()}
          if all(len(i) >= 2 for i in results.values()):
            # ====== print text plot ====== #
            if self.print_plot:
              text = []
              for t_name in task_name:
                values = results[t_name]
                if isinstance(values[0], Number):
                  t = V.print_bar(f=values, height=8,
                                  title=t_name + "/" + o_name)
                elif isinstance(values[0], np.ndarray) and values[0].ndim == 2 and \
                values[0].shape[0] == values[0].shape[1]:
                  t = V.print_confusion(arr=values[-1],
                                       side_bar=False, inc_stats=True,
                                       float_precision=2)
                else:
                  t = ''
                if len(t) > 0:
                  text.append(t)
              if len(text) > 1:
                print(V.merge_text_graph(*text, padding='  '))
              else:
                print(text[0])
            # ====== matplotlib plot and save pdf ====== #
            if self.save_path is not None:
              for t_idx, t_name in enumerate(task_name):
                values = results[t_name]
                # plotting series
                if isinstance(values[0], Number):
                  if not save_figures:
                    V.plot_figure(nrow=int(n_row * 1.8), ncol=20)
                  save_figures = True

                  max_epoch = np.argmax(values)
                  max_val = values[max_epoch]
                  min_epoch = np.argmin(values)
                  min_val = values[min_epoch]

                  plt.subplot(n_row, n_col, o_idx * len(task_name) + t_idx + 1)
                  plt.plot(values)
                  plt.scatter(max_epoch, max_val, s=180, alpha=0.4, c='r')
                  plt.scatter(min_epoch, min_val, s=180, alpha=0.4, c='g')

                  plt.xlim((0, len(values) - 1))
                  if not np.any(np.isinf(values)):
                    eps = 0.1 * (max_val - min_val)
                    plt.ylim((min_val - eps, max_val + eps))
                  plt.xticks(np.linspace(0, len(values) - 1, num=12,
                                         dtype='int32'))

                  title_text = '[%s]' % o_name if t_idx == 0 else ''
                  title_text += t_name
                  plt.title('%s' % title_text,
                            fontsize=8, fontweight='bold')
        # save figure to pdf or image files
        if save_figures:
          if override:
            save_path = self.save_path
          else:
            path, ext = os.path.splitext(self.save_path)
            save_path = path + ('.%d' % (task.curr_epoch + 1)) + ext
          V.plot_save(save_path, tight_plot=True,
                      clear_all=True, log=False, dpi=180)
          self.send_notification("Saved summary at: %s" % save_path)
    return None



import os
import shutil
from collections import OrderedDict

import numpy as np
import tensorflow as tf

from odin import backend as K
from odin.utils import (is_string, is_number, ctext, wprint,
                        is_primitives, is_pickleable, FuncDesc)

# ===========================================================================
# Helper
# ===========================================================================
def _check_dtype(dtype):
  if hasattr(dtype, '__call__'):
    return dtype
  # ====== check dtype ====== #
  if dtype is None:
    dtype = K.floatX
  elif isinstance(dtype, np.dtype) or is_string(dtype):
    dtype = str(dtype)
  elif isinstance(dtype, VariableDesc):
    dtype = dtype.dtype
  elif isinstance(dtype, tf.DType):
    dtype = dtype.base_dtype.name
  return dtype

def _check_shape(s):
  if hasattr(s, '__call__'):
    return s
  if is_number(s) or s is None:
    s = (s,)
  elif isinstance(s, np.ndarray):
    s = s.tolist()
  return tuple([int(i) if is_number(i) else None for i in s])

def _shape_compare(shape1, shape2):
  """Return True if shape1 == shape2"""
  if len(shape1) != len(shape2): # different ndim
    return False
  for s1, s2 in zip(shape1, shape2):
    if s1 is None or s2 is None or s1 == -1 or s2 == -1:
      continue
    if s1 != s2:
      return False
  return True

# ===========================================================================
# Main Descriptor
# ===========================================================================
class VariableDesc(object):
  """ VariableDesc
  Store all the necessary information to create placeholder as input
  to any ComputationalGraph.

  Parameters
  ----------
  shape: tuple, list, TensorVariable, call-able
      if TensorVariable is given, shape and dtype will be taken from
      given variable. if a call-able object is given, the object must
      return shape information when called without any argument.
  dtype: str, numpy.dtype, call-able
      dtype of input variable
  name: str, None, call-able
      specific name for the variable

  Note
  ----
  This object is pickle-able and comparable
  """

  def __init__(self, shape, dtype=None, name=None):
    super(VariableDesc, self).__init__()
    # ====== placeholder ====== #
    self.__placeholder = None
    self._name = name if name is None else str(name)
    # Given a TensorVariabe, we don't want to pickle TensorVariable,
    # so copy all necessary information
    if K.is_tensor(shape):
      if dtype is None:
        self._dtype = _check_dtype(shape.dtype)
      self._shape = shape.shape.as_list()
      # store the placeholder so don't have to create it again
      self.__placeholder = shape
    # input the VariableDesc directly
    elif isinstance(shape, VariableDesc):
      self._shape = shape.shape
      self._dtype = shape.dtype if dtype is None \
          else _check_dtype(dtype)
      if shape.__placeholder is not None:
        self.__placeholder = shape.__placeholder
    # input regular information flow
    else:
      self._shape = _check_shape(shape)
      self._dtype = _check_dtype(dtype)

  # ==================== pickle ==================== #
  def __getstate__(self):
    return (self._shape, self._dtype, self._name)

  def __setstate__(self, states):
    (self._shape, self._dtype, self._name) = states
    self.__placeholder = None

  # ==================== properties ==================== #
  def set_placeholder(self, plh):
    if not K.is_placeholder(plh):
      raise ValueError("a placholder must be specified.")
    if plh.shape.as_list() == self.shape and \
    _check_dtype(plh.dtype) == self.dtype:
      self.__placeholder = plh
    else:
      raise ValueError("This VariableDesc require input with shape=%s,"
                       "and dtype=%s, but given a placholder with shape=%s, "
                       "dtype=%s." % (str(self.shape), self.dtype,
                      str(plh.shape.as_list()), _check_dtype(plh.dtype)))
    return self

  @property
  def placeholder(self):
    if self.__placeholder is None:
      self.__placeholder = K.placeholder(
          shape=self.shape, dtype=self.dtype, name=self.name)
    return self.__placeholder

  @property
  def name(self):
    return self._name

  @property
  def shape(self):
    s = self._shape() if hasattr(self._shape, '__call__') \
        else self._shape
    return tuple(s)

  @property
  def dtype(self):
    return self._dtype() if hasattr(self._dtype, '__call__') \
        else self._dtype

  # ==================== override ==================== #
  def __str__(self):
    return "<%s - name:%s shape:%s dtype:%s init:%s>" % \
    (ctext('VarDesc', 'cyan'), ctext(str(self.name), 'yellow'),
        str(self.shape), str(self.dtype),
     False if self.__placeholder is None else True)

  def __repr__(self):
    return self.__str__()

  def is_equal(self, other):
    # ====== compare to a TensorVariable ====== #
    if K.is_tensor(other):
      other = VariableDesc(
          shape=other.shape.as_list(),
          dtype=_check_dtype(other.dtype))
    # ====== compare to a VariableDesc ====== #
    if isinstance(other, VariableDesc):
      if _shape_compare(self.shape, other.shape) \
      and self.dtype == other.dtype:
        return True
    # ====== compare to a shape tuple (ignore the dtype) ====== #
    elif isinstance(other, (tuple, list)):
      return True if _shape_compare(self.shape, other) else False
    return False

class NNOpDesc(object):

  def __init__(self, nnop_type):
    super(NNOpDesc, self).__init__()
    assert isinstance(nnop_type, type)
    self._nnop_type = nnop_type
    self._save_states = {}

    # mapping: name -> VariableDesc, or Primitives
    self._kwargs_desc = OrderedDict()
    # mapping: variable_name -> (tensorflow_name, 'tensor' or 'variable')
    self._variable_info = OrderedDict()

    # this is special tricks, the unpickled ops stay useless
    # until its variables are restored, but if we restore the
    # variable right away, it create a session and prevent
    # any possibility of running tensorflow with multiprocessing
    # => store the _restore_vars_path for later, and restore
    # the variable when the NNOp is actually in used.
    self._set_restore_info(None, False)

  def _set_restore_info(self, vars_path, delete_after):
    self._restore_vars_path = vars_path
    self._delete_vars_folder = bool(delete_after)
    return self

  def _restore_variables(self):
    """ This method can be called anywhere to make sure
    the variable related to this NNOp is restored after
    pickling.
    """
    if hasattr(self, '_restore_vars_path') and \
    self._restore_vars_path is not None:
      folder_path = os.path.dirname(self._restore_vars_path)
      if os.path.exists(folder_path):
        K.restore_variables(self._restore_vars_path)
        # delete cached folder if necessary
        if self._delete_vars_folder:
          shutil.rmtree(folder_path)
      else:
        wprint("NNOp: '%s' cannot restore variables from path: '%s'"
               (self.name, folder_path))
      # reset info
      self._set_restore_info(None, False)

  def __setattr__(self, name, value):
    # this record all assigned attribute to pickle them later
    # check hasattr to prevent recursive loop at the beginning before
    # __init__ is called
    if hasattr(self, '_save_states'):
      if name not in ('_save_states',):
        if is_primitives(value, inc_ndarray=True,
                         exception_types=[NNOpDesc, FuncDesc]) or \
        (hasattr(value, '__call__') and is_pickleable(value)):
          self._save_states[name] = value
    return super(NNOpDesc, self).__setattr__(name, value)




from six import add_metaclass
from abc import ABCMeta, abstractproperty

from odin import backend as K
from odin.utils import flatten_list
from odin.nnet.base import Dense
from odin.nnet.models.base import Model

@add_metaclass(ABCMeta)
class _BNFbase(Model):

  @abstractproperty
  def nb_layers(self):
    raise NotImplementedError

  def get_input_info(self):
    w0 = self.get_loaded_param('w0')
    nb_ceps = w0.shape[1]
    return {'X': ((None, nb_ceps), 'float32')}

  def _initialize(self):
    param_names = flatten_list([('b%d' % i, 'w%d' % i)
                                for i in range(self.nb_layers)])
    weights = self.get_loaded_param(param_names)
    # ====== create ====== #
    layers = []
    for i in range(self.nb_layers):
      b = weights[i * 2].ravel()
      W = weights[i * 2 + 1].T
      num_units = b.shape[0]
      if i == self.nb_layers - 1:
        name = 'Bottleneck'
        nonlinearity = K.linear
      else:
        name = "Layer%d" % (i + 1)
        nonlinearity = K.relu
      op = Dense(num_units=num_units, W_init=W, b_init=b,
                 activation=nonlinearity, name=name)
      layers.append(op)
    self.layers = layers

  def _apply(self, X):
    # ====== apply ====== #
    for op in self.layers[:-1]:
      X = op(X)
      X = K.renorm_rms(X, axis=1)
    return self.layers[-1](X)


class BNF_1024_MFCC39(_BNFbase):
  """ Bottleneck fully connected network (1024-units):
  + Feature: 39-D MFCCs with 13-D+detla+ddelta)
     - sample_rate = 8000
     - filter_lo_edge = 100
     - filter_hi_edge = 4000
     - num_cepstral_coefs = 13
     - frame_length = 0.025
     - frame_shift = 0.010
     - preemphasis_coef = 0.97
     - num_channels = 40
     - num_fft_points = 512
     - window_type = hamm
     - spectrum_type = mag
     - compression_type = log
     - NO RASTA filter applied on MFCC
  + Input features must be normalized: (x - mean) / std
  + Context size: 21
  + Nonlinearity: Relu
  + Renorm: True (scales the data such that RMS is 1.0,
           performed after the activation)
  NOTE: the last layer (the bottleneck) is linear activated, and no renorm.
  """

  @property
  def nb_layers(self):
    return 5


class BNF_2048_MFCC39(_BNFbase):
  """ Bottleneck fully connected network (2048-units):
  + Feature: 39-D MFCCs with 13-D+detla+ddelta)
     - sample_rate = 8000
     - filter_lo_edge = 100
     - filter_hi_edge = 4000
     - num_cepstral_coefs = 13
     - frame_length = 0.025
     - frame_shift = 0.010
     - preemphasis_coef = 0.97
     - num_channels = 40
     - num_fft_points = 512
     - window_type = hamm
     - spectrum_type = mag
     - compression_type = log
     - NO RASTA filter applied on MFCC
  + Input features must be normalized: (x - mean) / std
  + Context size: 21
  + Nonlinearity: Relu
  + Renorm: True (scales the data such that RMS is 1.0,
           performed after the activation)
  NOTE: the last layer (the bottleneck) is linear activated, and no renorm.
  """
  @property
  def nb_layers(self):
    return 6


class BNF_2048_MFCC40(_BNFbase):
  """ Bottleneck fully connected network (2048-units):
    + Feature: static 40-D MFCCs
       - sample_rate = 8000
       - filter_lo_edge = 100
       - filter_hi_edge = 4000
       - num_cepstral_coefs = 40
       - frame_length = 0.025
       - frame_shift = 0.010
       - preemphasis_coef = 0.97
       - num_channels = 40
       - num_fft_points = 512
       - window_type = hamm
       - spectrum_type = mag
       - compression_type = log
       - NO RASTA filter applied on MFCC
    + Input features must be normalized: (x - mean) / std
    + Context size: 21
    + Nonlinearity: Relu
    + Renorm: True (scales the data such that RMS is 1.0,
              performed after the activation)
   NOTE: the last layer (the bottleneck) is linear activated, and no renorm.
  """
  @property
  def nb_layers(self):
    return 6



from odin.nnet.models.base import Model

import numpy as np
import tensorflow as tf


class InceptionV3(Model):
  """ One way to perform transfer learning is to remove the
  final classification layer of the network and extract the
  next-to-last layer of the CNN, in this case a 2048
  dimensional vector. """

  def __init__(self, offset_top=0, **kwargs):
    super(InceptionV3, self).__init__(**kwargs)

  def get_input_info(self):
    pass



from __future__ import print_function, division, absolute_import


from odin import backend as K
from odin.nnet.base import NNOp
from odin.utils import as_tuple
from odin.backend.role import (BatchNormPopulationMean, BatchNormScaleParameter,
                        BatchNormPopulationInvStd, BatchNormShiftParameter)

import tensorflow as tf
from tensorflow.python.ops import init_ops

class BatchNorm(NNOp):
  """ This class is adpated from Lasagne:
  Original work Copyright (c) 2014-2015 lasagne contributors
  All rights reserved.
  LICENSE: https://github.com/Lasagne/Lasagne/blob/master/LICENSE

  Batch Normalization

  This layer implements batch normalization of its inputs, following [1]_:

  .. math::
      y = \\frac{x - \\mu}{\\sqrt{\\sigma^2 + \\epsilon}} \\gamma + \\beta

  That is, the input is normalized to zero mean and unit variance, and then
  linearly transformed. The crucial part is that the mean and variance are
  computed across the batch dimension, i.e., over examples, not per example.

  During training, :math:`\\mu` and :math:`\\sigma^2` are defined to be the
  mean and variance of the current input mini-batch :math:`x`, and during
  testing, they are replaced with average statistics over the training
  data. Consequently, this layer has four stored parameters: :math:`\\beta`,
  :math:`\\gamma`, and the averages :math:`\\mu` and :math:`\\sigma^2`
  (nota bene: instead of :math:`\\sigma^2`, the layer actually stores
  :math:`1 / \\sqrt{\\sigma^2 + \\epsilon}`, for compatibility to cuDNN).
  By default, this layer learns the average statistics as exponential moving
  averages computed during training, so it can be plugged into an existing
  network without any changes of the training procedure (see Notes).

  Parameters
  ----------
  axes : 'auto', int or tuple of int
      The axis or axes to normalize over. If ``'auto'`` (the default),
      normalize over all axes except for the final (often used to represent
      channel or feature map).
  epsilon : scalar
      Small constant :math:`\\epsilon` added to the variance before taking
      the square root and dividing by it, to avoid numerical problems
  alpha : scalar
      Coefficient for the exponential moving average of batch-wise means and
      standard deviations computed during training; the closer to one, the
      more it will depend on the last batches seen
  beta : trainable variable, expression, numpy array, call-able or None
      Initial value, expression or initializer for :math:`\\beta`. Must match
      the incoming shape, skipping all axes in `axes`. Set to ``None`` to fix
      it to 0.0 instead of learning it.
      See :func:`lasagne.utils.create_param` for more information.
  gamma : trainable variable, expression, numpy array, call-able or None
      Initial value, expression or initializer for :math:`\\gamma`. Must
      match the incoming shape, skipping all axes in `axes`. Set to ``None``
      to fix it to 1.0 instead of learning it.
      See :func:`lasagne.utils.create_param` for more information.
  mean : Theano shared variable, expression, numpy array, or call-able
      Initial value, expression or initializer for :math:`\\mu`. Must match
      the incoming shape, skipping all axes in `axes`.
      See :func:`lasagne.utils.create_param` for more information.
  inv_std : Theano shared variable, expression, numpy array, or call-able
      Initial value, expression or initializer for :math:`1 / \\sqrt{
      \\sigma^2 + \\epsilon}`. Must match the incoming shape, skipping all
      axes in `axes`.
      See :func:`lasagne.utils.create_param` for more information.
  noise_level : None, float or tensor scalar
      if `noise_level` is not None, it specify standard deviation of
      added Gaussian noise. The noise will be applied before adding `beta`.
  noise_dims: int, list(int), or 'auto'
      these dimensions will be setted to 1 in noise_shape, and
      used to broadcast the dropout mask.
      If `noise_dims` is "auto", it will be the same as `axes`
  fused: bool
      if `True`, use a faster, fused implementation based on
      nn.fused_batch_norm. If `None`, use the fused implementation if possible.
  **kwargs
      Any additional keyword arguments are passed to the :class:`Layer`
      superclass.

  Notes
  -----
  This layer should be inserted between a linear transformation (such as a
  :class:`DenseLayer`, or :class:`Conv2DLayer`) and its activation. The
  convenience function :func:`batch_norm` modifies an existing layer to
  insert batch normalization in front of its activation.

  The behavior can be controlled by passing keyword arguments to
  :func:`lasagne.layers.get_output()` when building the output expression
  of any network containing this layer.

  During training, [1]_ normalize each input mini-batch by its statistics
  and update an exponential moving average of the statistics to be used for
  validation. This can be achieved by passing ``deterministic=False``.
  For validation, [1]_ normalize each input mini-batch by the stored
  statistics. This can be achieved by passing ``deterministic=True``.

  For more fine-grained control, ``batch_norm_update_averages`` can be passed
  to update the exponential moving averages (``True``) or not (``False``),
  and ``batch_norm_use_averages`` can be passed to use the exponential moving
  averages for normalization (``True``) or normalize each mini-batch by its
  own statistics (``False``). These settings override ``deterministic``.

  Note that for testing a model after training, [1]_ replace the stored
  exponential moving average statistics by fixing all network weights and
  re-computing average statistics over the training data in a layerwise
  fashion. This is not part of the layer implementation.

  In case you set `axes` to not include the batch dimension (the first axis,
  usually), normalization is done per example, not across examples. This does
  not require any averages, so you can pass ``batch_norm_update_averages``
  and ``batch_norm_use_averages`` as ``False`` in this case.

  Example
  --------
  For convolution output with the shape
  (nb_samples, nb_channel, width, height) you can normalize axes=(0, 1, 2, 3)
  or any subset of above axes, but you must include the first dimension (0,)
  because we don't know the batch size in advance.


  References
  ----------
  .. [1] Ioffe, Sergey and Szegedy, Christian (2015):
         Batch Normalization: Accelerating Deep Network Training by Reducing
         Internal Covariate Shift. http://arxiv.org/abs/1502.03167.
  """

  def __init__(self, axes='auto', epsilon=1e-4, alpha=0.1,
               beta_init=init_ops.constant_initializer(0),
               gamma_init=init_ops.constant_initializer(1),
               mean_init=init_ops.constant_initializer(0),
               var_init=init_ops.constant_initializer(1),
               noise_level=None, noise_dims='auto',
               activation=K.linear, **kwargs):
    super(BatchNorm, self).__init__(**kwargs)
    self.axes = axes
    self.epsilon = epsilon
    self.alpha = alpha
    self.beta_init = beta_init
    self.gamma_init = gamma_init
    self.mean_init = mean_init
    self.var_init = var_init
    # ====== noise ====== #
    self.noise_level = noise_level
    self.noise_dims = noise_dims
    self.activation = K.linear if activation is None else activation

  # ==================== abstract method ==================== #
  def _initialize(self):
    if self.axes == 'auto':
      # default: normalize over all but the second axis
      # for so-called "global normalization", used with convolutional
      # filters with shape [batch, height, width, depth],
      # pass axes=[0, 1, 2]
      self.axes = tuple(range(0, len(self.input_shape) - 1))
    else:
      self.axes = as_tuple(self.axes, t=int)
    # check noise_dims
    if self.noise_dims == 'auto':
      self.noise_dims = self.axes
    # create parameters, ignoring all dimensions provided in axes
    shape = [size for axis, size in enumerate(self.input_shape)
             if axis not in self.axes]
    if any(size is None for size in shape):
      raise ValueError("BatchNorm needs specified input sizes for "
                       "all axes not normalized over.")
    # init learnable parameters
    if self.beta_init is not None:
      self.get_variable_nnop(initializer=self.beta_init,
          shape=shape, name='beta',
          roles=BatchNormShiftParameter)
    if self.gamma_init is not None:
      self.get_variable_nnop(initializer=self.gamma_init,
          shape=shape, name='gamma',
          roles=BatchNormScaleParameter)
    # running mean and invert std
    if self.mean_init is not None:
      self.get_variable_nnop(initializer=self.mean_init,
          shape=shape, name='mean',
          roles=BatchNormPopulationMean)
    if self.var_init is not None:
      self.get_variable_nnop(initializer=self.var_init,
          shape=shape, name='var',
          roles=BatchNormPopulationInvStd)

  def _apply(self, X, noise=0):
    ndim = X.shape.ndims
    # if is training, normalize input by its own mean and std
    mean, var = tf.nn.moments(X, axes=self.axes)
    # prepare dimshuffle pattern inserting broadcastable axes as needed
    param_axes = iter(range(ndim - len(self.axes)))
    pattern = ['x' if input_axis in self.axes else next(param_axes)
               for input_axis in range(ndim)]
    # apply dimshuffle pattern to all parameters
    beta = 0 if self.beta_init is None else \
        K.dimshuffle(self.get('beta'), pattern)
    gamma = 1 if self.gamma_init is None else \
        K.dimshuffle(self.get('gamma'), pattern)

    # ====== if trainign: use local mean and var ====== #
    def training_fn():
      running_mean = ((1 - self.alpha) * self.get('mean') +
                      self.alpha * mean)
      running_var = ((1 - self.alpha) * self.get('var') +
                     self.alpha * var)
      with tf.control_dependencies([
              tf.assign(self.get('mean'), running_mean),
              tf.assign(self.get('var'), running_var)]):
        return tf.identity(mean), tf.identity(var)

    # ====== if inference: use global mean and var ====== #
    def infer_fn():
      return self.get('mean'), self.get('var')

    mean, var = tf.cond(K.is_training(), training_fn, infer_fn)
    inv_std = tf.rsqrt(var + self.epsilon)
    normalized = (X - K.dimshuffle(mean, pattern)) * \
        (gamma * K.dimshuffle(inv_std, pattern))
    # ====== applying noise if required ====== #
    if self.noise_level is not None:
      normalized = K.rand.apply_noise(normalized,
          level=self.noise_level, noise_dims=self.noise_dims,
          noise_type='gaussian')
    # add beta
    normalized = normalized + beta
    # activated output
    return self.activation(normalized)
